from PyQt5.QtWidgets import QHBoxLayout, QLabel, QWidget, QDoubleSpinBox, QAbstractSpinBox


def labeled_spinbox(name, default_value, lb, ub, decimals=4):
    widget = QWidget()
    hbox = QHBoxLayout(widget)
    hbox.setContentsMargins(0, 0, 0, 0)
    label = QLabel(name)
    widget.sb = sb = QDoubleSpinBox()
    sb.setDecimals(decimals)

    # Sanity check
    if lb is None:
        lb = -1e3
    if ub is None:
        ub = 1e3
    if default_value is None:
        default_value = 0

    sb.setRange(lb, ub)
    sb.setStepType(QAbstractSpinBox.AdaptiveDecimalStepType)
    sb.setValue(default_value)
    hbox.addWidget(label)
    hbox.addWidget(sb, 1)

    return widget



from PyQt5.QtCore import Qt


def test_gui_main(qtbot):
    from badger.gui.windows.main_window import BadgerMainWindow

    window = BadgerMainWindow()
    window.show()
    qtbot.addWidget(window)

    # Test new routine feature
    qtbot.mouseClick(window.home_page.btn_new, Qt.MouseButton.LeftButton)
    assert window.stacks.currentIndex() == 1



import logging
logger = logging.getLogger(__name__)
from ..utils import yprint


def show_algo(args):
    try:
        from ..factory import list_algo, get_algo
    except Exception as e:
        logger.error(e)
        return

    if args.algo_name is None:
        yprint(list_algo())
        return

    try:
        _, configs = get_algo(args.algo_name)
        yprint(configs)
    except Exception as e:
        logger.error(e)
        try:
            # The exception could carry the configs information
            configs = e.configs
            yprint(configs)
        except:
            pass



import numpy as np
from PyQt5.QtWidgets import QVBoxLayout, QHBoxLayout, QWidget, QPushButton
from PyQt5.QtWidgets import QMessageBox, QComboBox, QLabel, QStyledItemDelegate
from PyQt5.QtCore import pyqtSignal, QThreadPool
from PyQt5.QtGui import QFont
import pyqtgraph as pg
from .routine_runner import BadgerRoutineRunner
# from ...utils import AURORA_PALETTE, FROST_PALETTE
from ...utils import norm, ParetoFront
from ...logbook import send_to_logbook, BADGER_LOGBOOK_ROOT
from ...archive import archive_run, BADGER_ARCHIVE_ROOT


stylesheet_del = '''
QPushButton:hover:pressed
{
    background-color: #C7737B;
}
QPushButton:hover
{
    background-color: #BF616A;
}
QPushButton
{
    background-color: #A9444E;
}
'''

class BadgerOptMonitor(QWidget):
    sig_pause = pyqtSignal(bool)  # True: pause, False: resume
    sig_stop = pyqtSignal()
    sig_lock = pyqtSignal(bool)  # True: lock GUI, False: unlock GUI
    sig_new_run = pyqtSignal()
    sig_run_name = pyqtSignal(str)  # filename of the new run
    sig_inspect = pyqtSignal(int)  # index of the inspector
    sig_progress = pyqtSignal(list, list, list)  # new evaluated solution
    sig_del = pyqtSignal()

    def __init__(self):
        super().__init__()
        # self.setAttribute(Qt.WA_DeleteOnClose, True)

        # For plot type switching
        self.x_plot_type = 0  # 0: raw, 1: normalized
        # Routine info
        self.routine = None
        self.var_names = []
        self.obj_names = []
        self.con_names = []
        self.vranges = []
        # Curves in the monitor
        self.curves_var = []
        self.curves_obj = []
        self.curves_con = []
        # Data to be visualized
        self.vars = []
        self.objs = []
        self.cons = []
        # Run optimization
        self.thread_pool = None
        self.routine_runner = None
        self.running = False
        # Analysis tool for history runs
        self.pf = None

        self.init_ui()
        self.config_logic()

    def init_ui(self):
        # self.main_panel = main_panel = QWidget(self)
        # main_panel.setStyleSheet('background-color: #19232D;')
        vbox = QVBoxLayout(self)
        vbox.setContentsMargins(0, 0, 0, 0)

        # Config bar
        config_bar = QWidget()
        hbox_config = QHBoxLayout(config_bar)
        hbox_config.setContentsMargins(0, 0, 0, 0)
        label = QLabel('Evaluation History (X) Plot Type')
        self.cb_plot = cb_plot = QComboBox()
        cb_plot.setItemDelegate(QStyledItemDelegate())
        cb_plot.addItems(['Raw', 'Normalized'])
        hbox_config.addWidget(label)
        hbox_config.addWidget(cb_plot, 1)

        # Set up the monitor
        # Don't set show=True or there will be a blank window flashing once
        self.monitor = monitor = pg.GraphicsLayoutWidget()
        pg.setConfigOptions(antialias=True)
        # monitor.ci.setBorder((50, 50, 100))
        # monitor.resize(1000, 600)

        self.plot_obj = plot_obj = monitor.addPlot(
            title='Evaluation History (Y)')
        plot_obj.setLabel('left', 'objectives')
        plot_obj.setLabel('bottom', 'iterations')
        plot_obj.showGrid(x=True, y=True)
        leg_obj = plot_obj.addLegend()
        leg_obj.setBrush((50, 50, 100, 200))

        monitor.nextRow()  # leave space for the cons plot
        monitor.nextRow()

        self.plot_var = plot_var = monitor.addPlot(
            title='Evaluation History (X)')
        plot_var.setLabel('left', 'variables')
        plot_var.setLabel('bottom', 'iterations')
        plot_var.showGrid(x=True, y=True)
        leg_var = plot_var.addLegend()
        leg_var.setBrush((50, 50, 100, 200))

        plot_var.setXLink(plot_obj)

        self.colors = ['c', 'g', 'm', 'y', 'b', 'r', 'w']
        self.symbols = ['o', 't', 't1', 's', 'p', 'h', 'd']

        self.ins_obj = pg.InfiniteLine(movable=True, angle=90, label=None,
                                       labelOpts={
                                           'position': 0.1,
                                           'color': (200, 200, 100),
                                           'fill': (200, 200, 200, 50),
                                           'movable': True})
        self.ins_con = pg.InfiniteLine(movable=True, angle=90, label=None,
                                       labelOpts={
                                           'position': 0.1,
                                           'color': (200, 200, 100),
                                           'fill': (200, 200, 200, 50),
                                           'movable': True})
        self.ins_var = pg.InfiniteLine(movable=True, angle=90, label=None,
                                       labelOpts={
                                           'position': 0.1,
                                           'color': (200, 200, 100),
                                           'fill': (200, 200, 200, 50),
                                           'movable': True})
        plot_obj.addItem(self.ins_obj)
        plot_var.addItem(self.ins_var)

        # Action bar
        action_bar = QWidget()
        # action_bar.hide()
        hbox_action = QHBoxLayout(action_bar)
        hbox_action.setContentsMargins(0, 0, 0, 0)

        cool_font = QFont()
        cool_font.setWeight(QFont.DemiBold)
        # cool_font.setPixelSize(16)

        self.btn_del = btn_del = QPushButton('Delete Run')
        btn_del.setFixedSize(128, 32)
        btn_del.setFont(cool_font)
        btn_del.setStyleSheet(stylesheet_del)
        # self.btn_edit = btn_edit = QPushButton('Edit')
        # btn_edit.setFixedSize(64, 32)
        # btn_edit.setFont(cool_font)
        self.btn_log = btn_log = QPushButton('Logbook')
        btn_log.setFixedSize(64, 32)
        btn_log.setFont(cool_font)
        self.btn_reset = btn_reset = QPushButton('Reset')
        btn_reset.setDisabled(True)
        btn_reset.setFixedSize(64, 32)
        btn_reset.setFont(cool_font)
        self.btn_opt = btn_opt = QPushButton('Optimal')
        btn_opt.setFixedSize(64, 32)
        btn_opt.setFont(cool_font)
        self.btn_set = btn_set = QPushButton('Set')
        btn_set.setDisabled(True)
        btn_set.setFixedSize(64, 32)
        btn_set.setFont(cool_font)
        self.btn_ctrl = btn_ctrl = QPushButton('Pause')
        btn_ctrl.setDisabled(True)
        btn_ctrl.setFixedSize(64, 32)
        btn_ctrl.setFont(cool_font)
        self.btn_stop = btn_stop = QPushButton('Run')
        btn_stop.setDisabled(True)
        btn_stop.setFixedSize(128, 32)
        btn_stop.setFont(cool_font)
        hbox_action.addWidget(btn_del)
        # hbox_action.addWidget(btn_edit)
        hbox_action.addWidget(btn_log)
        hbox_action.addWidget(btn_opt)
        hbox_action.addStretch(1)
        hbox_action.addWidget(btn_reset)
        hbox_action.addWidget(btn_set)
        hbox_action.addWidget(btn_ctrl)
        hbox_action.addWidget(btn_stop)

        vbox.addWidget(config_bar)
        vbox.addWidget(monitor)
        vbox.addWidget(action_bar)

    def config_logic(self):
        # Sync the inspector lines
        self.ins_obj.sigDragged.connect(self.ins_obj_dragged)
        self.ins_obj.sigPositionChangeFinished.connect(self.ins_drag_done)
        self.ins_con.sigDragged.connect(self.ins_con_dragged)
        self.ins_con.sigPositionChangeFinished.connect(self.ins_drag_done)
        self.ins_var.sigDragged.connect(self.ins_var_dragged)
        self.ins_var.sigPositionChangeFinished.connect(self.ins_drag_done)
        self.plot_obj.scene().sigMouseClicked.connect(self.on_mouse_click)
        # sigMouseReleased.connect(self.on_mouse_click)

        # Thread runner
        self.thread_pool = QThreadPool(self)

        self.btn_del.clicked.connect(self.delete_run)
        self.btn_log.clicked.connect(self.logbook)
        self.btn_reset.clicked.connect(self.reset_env)
        self.btn_opt.clicked.connect(self.jump_to_optimal)
        self.btn_set.clicked.connect(self.set_vars)
        self.btn_ctrl.clicked.connect(self.ctrl_routine)
        self.btn_stop.clicked.connect(self.run_stop_routine)

        # Visualization
        self.cb_plot.currentIndexChanged.connect(self.select_x_plot_type)

    def init_plots(self, routine, data=None, run_filename=None):
        # Parse routine
        self.routine = routine
        try:
            self.obj_names = [next(iter(d))
                              for d in self.routine['config']['objectives']]
        except:
            self.obj_names = []
        try:
            self.var_names = [next(iter(d))
                              for d in self.routine['config']['variables']]
        except:
            self.var_names = []
        try:
            self.vranges = np.array([d[next(iter(d))]
                                    for d in routine['config']['variables']])
        except:
            self.vranges = []
        try:
            if self.routine['config']['constraints']:
                self.con_names = [next(iter(d))
                                  for d in self.routine['config']['constraints']]
            else:
                self.con_names = []
        except:
            self.con_names = []

        # Configure plots
        # Clear current plots
        self.plot_obj.clear()
        self.plot_obj.addItem(self.ins_obj)
        try:
            self.plot_con.clear()
            self.plot_con.addItem(self.ins_con)
        except:
            pass
        self.plot_var.clear()
        self.plot_var.addItem(self.ins_var)
        # Put in the empty curves
        self.curves_var = []
        self.curves_obj = []
        self.curves_con = []

        for i, obj_name in enumerate(self.obj_names):
            color = self.colors[i % len(self.colors)]
            symbol = self.symbols[i % len(self.colors)]
            _curve = self.plot_obj.plot(pen=pg.mkPen(color, width=3),
                                        # symbol=symbol,
                                        name=obj_name)
            self.curves_obj.append(_curve)

        for i, var_name in enumerate(self.var_names):
            color = self.colors[i % len(self.colors)]
            symbol = self.symbols[i % len(self.colors)]
            _curve = self.plot_var.plot(pen=pg.mkPen(color, width=3),
                                        # symbol=symbol,
                                        name=var_name)
            self.curves_var.append(_curve)

        if self.con_names:
            try:
                self.plot_con
            except:
                self.plot_con = plot_con = self.monitor.addPlot(
                    row=1, col=0, title='Evaluation History (C)')
                plot_con.setLabel('left', 'constraints')
                plot_con.setLabel('bottom', 'iterations')
                plot_con.showGrid(x=True, y=True)
                leg_con = plot_con.addLegend()
                leg_con.setBrush((50, 50, 100, 200))
                plot_con.addItem(self.ins_con)
                plot_con.setXLink(self.plot_obj)

            for i, con_name in enumerate(self.con_names):
                color = self.colors[i % len(self.colors)]
                symbol = self.symbols[i % len(self.colors)]
                _curve = self.plot_con.plot(pen=pg.mkPen(color, width=3),
                                            # symbol=symbol,
                                            name=con_name)
                self.curves_con.append(_curve)
        else:
            try:
                self.monitor.removeItem(self.plot_con)
                self.plot_con.removeItem(self.ins_con)
                del self.plot_con
            except:
                pass
        # Reset inspectors
        self.ins_obj.setValue(0)
        self.ins_var.setValue(0)
        self.ins_con.setValue(0)

        # Switch run button state
        if self.routine:
            self.btn_stop.setDisabled(False)
        else:
            self.btn_stop.setDisabled(True)

        # Fill in data
        self.data = data
        self.vars = []
        self.objs = []
        self.cons = []
        if self.routine_runner and self.routine_runner.run_filename == run_filename:
            self.btn_reset.setDisabled(False)
            self.btn_set.setDisabled(False)
        else:
            self.btn_reset.setDisabled(True)
            self.btn_set.setDisabled(True)
        if data is None:
            self.btn_del.setDisabled(True)
            self.btn_log.setDisabled(True)
            self.btn_opt.setDisabled(True)
            self.enable_auto_range()
            return

        for var in self.var_names:
            self.vars.append(data[var])
        self.vars = np.array(self.vars).T.tolist()
        for obj in self.obj_names:
            self.objs.append(data[obj])
        self.objs = np.array(self.objs).T.tolist()
        for con in self.con_names:
            self.cons.append(data[con])
        self.cons = np.array(self.cons).T.tolist()

        self.update_curves()

        self.calc_optimals()
        self.btn_del.setDisabled(False)
        self.btn_log.setDisabled(False)
        self.btn_opt.setDisabled(False)

    def init_routine_runner(self):
        self.reset_routine_runner()

        self.routine_runner = routine_runner = BadgerRoutineRunner(
            self.routine, False)
        routine_runner.signals.env_ready.connect(self.env_ready)
        routine_runner.signals.finished.connect(self.routine_finished)
        routine_runner.signals.progress.connect(self.update)
        routine_runner.signals.error.connect(self.on_error)
        routine_runner.signals.info.connect(self.on_info)

        self.sig_pause.connect(routine_runner.ctrl_routine)
        self.sig_stop.connect(routine_runner.stop_routine)

    def reset_routine_runner(self):
        if self.routine_runner:
            self.sig_pause.disconnect()
            self.sig_stop.disconnect()
            self.routine_runner = None

    def calc_optimals(self):
        rules = [d[next(iter(d))] for d in self.routine['config']['objectives']]
        self.pf = pf = ParetoFront(rules)

        for i, v in enumerate(self.vars):
            o = self.objs[i]
            idx = [i,] + v
            pf.is_dominated((idx, o))

    def start(self):
        self.sig_new_run.emit()
        self.init_plots(self.routine)
        self.init_routine_runner()
        self.running = True  # if a routine runner is working
        self.thread_pool.start(self.routine_runner)

        self.btn_stop.setText('Stop')
        self.btn_ctrl.setDisabled(False)
        self.sig_lock.emit(True)

    def is_critical(self, cons):
        if not self.con_names:
            return False, None

        constraints = self.routine['config']['constraints']
        for i, con_dict in enumerate(constraints):
            name = self.con_names[i]
            if len(con_dict[name]) != 3:
                continue

            value = cons[i]
            relation, thres = con_dict[name][:2]
            if relation == 'GREATER_THAN':
                if value <= thres:
                    return True, f'{name} (current value: {value:.4f}) is less than {thres}!'
            elif relation == 'LESS_THAN':
                if value >= thres:
                    return True, f'{name} (current value: {value:.4f}) is greater than {thres}!'
            else:
                if value != thres:
                    return True, f'{name} (current value: {value:.4f}) is not equal to {thres}!'

        return False, None

    def enable_auto_range(self):
        # Enable autorange
        self.plot_obj.enableAutoRange()
        self.plot_var.enableAutoRange()
        if self.con_names:
            self.plot_con.enableAutoRange()

    def update_curves(self):
        for i in range(len(self.obj_names)):
            self.curves_obj[i].setData(np.array(self.objs)[:, i])

        if self.con_names:
            for i in range(len(self.con_names)):
                self.curves_con[i].setData(np.array(self.cons)[:, i])

        if self.x_plot_type:
            _vars = norm(np.array(self.vars),
                         self.vranges[:, 0], self.vranges[:, 1])
        else:
            _vars = np.array(self.vars)
        for i in range(len(self.var_names)):
            self.curves_var[i].setData(_vars[:, i])

        self.enable_auto_range()

    def update(self, vars, objs, cons):
        self.vars.append(vars)
        self.objs.append(objs)
        self.cons.append(cons)

        self.update_curves()
        self.sig_progress.emit(vars, objs, cons)

        # Check critical condition
        critical, msg = self.is_critical(cons)
        if not critical:
            return

        self.sig_pause.emit(True)
        self.btn_ctrl.setText('Resume')

        reply = QMessageBox.warning(self,
                                    'Run Paused',
                                    f'Critical constraint was violated: {msg}\nTerminate the run?',
                                    QMessageBox.Yes | QMessageBox.No, QMessageBox.Yes)
        if reply == QMessageBox.Yes:
            self.sig_stop.emit()

    def env_ready(self, init_vars):
        self.env = self.routine_runner.env
        self.init_vars = init_vars

        self.btn_log.setDisabled(False)
        self.btn_opt.setDisabled(False)

    def routine_finished(self):
        self.running = False
        self.btn_ctrl.setText('Pause')
        self.btn_ctrl.setDisabled(True)
        self.btn_stop.setText('Run')
        self.btn_reset.setDisabled(False)
        self.btn_set.setDisabled(False)
        self.btn_del.setDisabled(False)

        self.sig_lock.emit(False)

        try:
            run = archive_run(self.routine_runner.routine, self.routine_runner.data)
            self.routine_runner.run_filename = run['filename']
            self.sig_run_name.emit(run['filename'])
        except Exception as e:
            QMessageBox.critical(self, 'Archive failed!', str(e))

        QMessageBox.information(
            self, 'Success!', f'Run data archived to {BADGER_ARCHIVE_ROOT}')

    def on_error(self, error):
        QMessageBox.critical(self, 'Error!', str(error))

    # Do not show info -- too distracting
    def on_info(self, msg):
        pass
        # QMessageBox.information(self, 'Info', msg)

    def logbook(self):
        try:
            if self.routine_runner:
                routine = self.routine_runner.routine
                data = self.routine_runner.data.to_dict('list')
            else:
                routine = self.routine
                data = self.data
            send_to_logbook(routine, data, self.monitor)
        except Exception as e:
            QMessageBox.critical(self, 'Log failed!', str(e))

        QMessageBox.information(
            self, 'Success!', f'Log saved to {BADGER_LOGBOOK_ROOT}')

    def ctrl_routine(self):
        if self.btn_ctrl.text() == 'Pause':
            self.sig_pause.emit(True)
            self.btn_ctrl.setText('Resume')
        else:
            self.sig_pause.emit(False)
            self.btn_ctrl.setText('Pause')

    def run_stop_routine(self):
        if self.btn_stop.text() == 'Run':
            self.start()
        else:
            self.sig_stop.emit()

    def ins_obj_dragged(self, ins_obj):
        self.ins_var.setValue(ins_obj.value())
        if self.con_names:
            self.ins_con.setValue(ins_obj.value())

    def ins_con_dragged(self, ins_con):
        self.ins_var.setValue(ins_con.value())
        self.ins_obj.setValue(ins_con.value())

    def ins_var_dragged(self, ins_var):
        self.ins_obj.setValue(ins_var.value())
        if self.con_names:
            self.ins_con.setValue(ins_var.value())

    def ins_drag_done(self, ins):
        value = np.round(ins.value())
        self.ins_obj.setValue(value)
        if self.con_names:
            self.ins_con.setValue(value)
        self.ins_var.setValue(value)

        self.sig_inspect.emit(value)

    def reset_env(self):
        reply = QMessageBox.question(self,
                                     'Reset Environment',
                                     f'Are you sure you want to reset the env vars back to {self.init_vars}?',
                                     QMessageBox.Yes | QMessageBox.No, QMessageBox.No)
        if reply != QMessageBox.Yes:
            return

        current_vars = self.env._get_vars(self.var_names)
        self.env._set_vars(self.var_names, self.init_vars)
        after_vars = self.env._get_vars(self.var_names)
        QMessageBox.information(
            self, 'Reset Environment', f'Env vars {current_vars} -> {after_vars}')

    def jump_to_optimal(self):
        try:
            pf = self.pf
        except:
            pf = self.routine_runner.pf
        if pf is None:
            return

        idx = pf.pareto_set[0][0]
        self.ins_obj.setValue(idx)
        if self.con_names:
            self.ins_con.setValue(idx)
        self.ins_var.setValue(idx)

        self.sig_inspect.emit(idx)

    def jump_to_solution(self, idx):
        self.ins_obj.setValue(idx)
        if self.con_names:
            self.ins_con.setValue(idx)
        self.ins_var.setValue(idx)

    def set_vars(self):
        df = self.routine_runner.data
        idx = int(self.ins_obj.value())
        solution = df.loc[idx, self.var_names].to_numpy()

        reply = QMessageBox.question(self,
                                     'Apply Solution',
                                     f'Are you sure you want to apply the current solution at {solution} to env?',
                                     QMessageBox.Yes | QMessageBox.No, QMessageBox.No)
        if reply != QMessageBox.Yes:
            return

        self.env._set_vars(self.var_names, solution)
        # center around the inspector
        self.plot_var.setXRange(idx - 3, idx + 3)
        # QMessageBox.information(
        #     self, 'Set Environment', f'Env vars have been set to {solution}')

    def select_x_plot_type(self, i):
        self.x_plot_type = i
        if i:
            _vars = norm(np.array(self.vars),
                         self.vranges[:, 0], self.vranges[:, 1])
        else:
            _vars = np.array(self.vars)
        for i in range(len(self.var_names)):
            self.curves_var[i].setData(_vars[:, i])

    def on_mouse_click(self, event):
        # https://stackoverflow.com/a/64081483
        coor_obj = self.plot_obj.vb.mapSceneToView(event._scenePos)
        if self.con_names:
            coor_con = self.plot_con.vb.mapSceneToView(event._scenePos)
        coor_var = self.plot_var.vb.mapSceneToView(event._scenePos)

        flag = self.plot_obj.viewRect().contains(coor_obj) or \
            self.plot_var.viewRect().contains(coor_var)
        if self.con_names:
            flag = flag or self.plot_con.viewRect().contains(coor_con)

        if flag:
            idx = int(np.round(coor_obj.x()))

            self.ins_obj.setValue(idx)
            if self.con_names:
                self.ins_con.setValue(idx)
            self.ins_var.setValue(idx)

            self.sig_inspect.emit(idx)

    def delete_run(self):
        self.sig_del.emit()

    # def closeEvent(self, event):
    #     if not self.running:
    #         return

    #     reply = QMessageBox.question(self,
    #                                  'Window Close',
    #                                  'Closing this window will terminate the run, and the run data would NOT be archived! Proceed?',
    #                                  QMessageBox.Yes | QMessageBox.No, QMessageBox.No)

    #     if reply == QMessageBox.Yes:
    #         self.sig_stop.emit()
    #         event.accept()
    #     else:
    #         event.ignore()



import os
import logging
logger = logging.getLogger(__name__)
from .db import save_run, remove_run_by_filename
from .utils import curr_ts_to_str, ystring, load_config
from .settings import read_value


# Check badger optimization run archive root
BADGER_ARCHIVE_ROOT = read_value('BADGER_ARCHIVE_ROOT')
if BADGER_ARCHIVE_ROOT is None:
    raise Exception('Please set the BADGER_ARCHIVE_ROOT env var!')
elif not os.path.exists(BADGER_ARCHIVE_ROOT):
    os.makedirs(BADGER_ARCHIVE_ROOT)
    logger.info(
        f'Badger run root {BADGER_ARCHIVE_ROOT} created')


def archive_run(routine, data):
    # routine: dict
    # data: pandas dataframe
    suffix = curr_ts_to_str("lcls-fname")
    tokens = suffix.split('-')
    first_level = tokens[0]
    second_level = f'{tokens[0]}-{tokens[1]}'
    third_level = f'{tokens[0]}-{tokens[1]}-{tokens[2]}'
    path = os.path.join(BADGER_ARCHIVE_ROOT, first_level, second_level, third_level)
    os.makedirs(path, exist_ok=True)
    fname = f'BadgerOpt-{suffix}.yaml'

    run = {
        'filename': fname,
        'routine': routine,
        'data': data.to_dict('list'),
    }
    rid = save_run(run)
    run = {'id': rid, **run}  # Put id in front

    with open(os.path.join(path, fname), 'w') as f:
        f.write(ystring(run))

    return run


def list_run():
    runs = {}
    # Get years, latest first
    years = sorted([p for p in os.listdir(BADGER_ARCHIVE_ROOT)
                   if os.path.isdir(os.path.join(BADGER_ARCHIVE_ROOT, p))], reverse=True)
    for year in years:
        path_year = os.path.join(BADGER_ARCHIVE_ROOT, year)
        months = sorted([p for p in os.listdir(path_year) if os.path.isdir(
            os.path.join(path_year, p))], reverse=True)
        runs[year] = {}
        for month in months:
            path_month = os.path.join(path_year, month)
            days = sorted([p for p in os.listdir(path_month) if os.path.isdir(
                os.path.join(path_month, p))], reverse=True)
            runs[year][month] = {}
            for day in days:
                path_day = os.path.join(path_month, day)
                files = [p for p in os.listdir(path_day) if os.path.splitext(p)[
                    1] == '.yaml']
                files = sorted(files, key=lambda f: os.path.getmtime(
                    os.path.join(path_day, f)), reverse=True)
                runs[year][month][day] = files

    return runs


def load_run(run_fname):
    tokens = run_fname.split('-')
    first_level = tokens[1]
    second_level = f'{tokens[1]}-{tokens[2]}'
    third_level = f'{tokens[1]}-{tokens[2]}-{tokens[3]}'

    return load_config(os.path.join(BADGER_ARCHIVE_ROOT, first_level, second_level, third_level, run_fname))


def delete_run(run_fname):
    tokens = run_fname.split('-')
    first_level = tokens[1]
    second_level = f'{tokens[1]}-{tokens[2]}'
    third_level = f'{tokens[1]}-{tokens[2]}-{tokens[3]}'

    # Remove record from the database
    remove_run_by_filename(run_fname)

    return os.remove(os.path.join(BADGER_ARCHIVE_ROOT, first_level, second_level, third_level, run_fname))



from PyQt5.QtWidgets import QApplication, QTableWidget, QTableWidgetItem, QHeaderView
from PyQt5.QtCore import Qt


# https://stackoverflow.com/questions/60715462/how-to-copy-and-paste-multiple-cells-in-qtablewidget-in-pyqt5
class TableWithCopy(QTableWidget):
    """
    this class extends QTableWidget
    * supports copying multiple cell's text onto the clipboard
    * formatted specifically to work with multiple-cell paste into programs
      like google sheets, excel, or numbers
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def keyPressEvent(self, event):
        super().keyPressEvent(event)
        if event.key() == Qt.Key.Key_C and (event.modifiers() & Qt.KeyboardModifier.ControlModifier):
            copied_cells = sorted(self.selectedIndexes())

            copy_text = ''
            max_column = copied_cells[-1].column()
            for c in copied_cells:
                copy_text += self.item(c.row(), c.column()).text()
                if c.column() == max_column:
                    copy_text += '\n'
                else:
                    copy_text += '\t'

            QApplication.clipboard().setText(copy_text)


def update_table(table, data=None):
    table.setRowCount(0)
    table.horizontalHeader().setVisible(False)

    if data is None:
        return table

    _data = data.copy()
    m = len(_data['timestamp_raw'])
    del _data['timestamp_raw']
    del _data['timestamp']
    n = len(_data)

    table.setRowCount(m)
    table.setColumnCount(n)
    for i, key in enumerate(_data.keys()):
        for j, v in enumerate(_data[key]):
            table.setItem(j, i, QTableWidgetItem(f'{v:g}'))
    table.setHorizontalHeaderLabels(list(_data.keys()))
    table.setVerticalHeaderLabels([str(i) for i in range(m)])  # row index starts from 0
    table.horizontalHeader().setVisible(True)

    return table


def reset_table(table, header):
    table.setRowCount(0)
    table.horizontalHeader().setVisible(False)
    table.setHorizontalHeaderLabels(header)
    table.horizontalHeader().setVisible(True)

    return table


def add_row(table, row):
    r = table.rowCount()
    table.insertRow(r)
    for i, v in enumerate(row):
        table.setItem(r, i, QTableWidgetItem(f'{v:g}'))
    table.setVerticalHeaderItem(r, QTableWidgetItem(str(r)))

    return table


def data_table(data=None):
    table = TableWithCopy()
    table.setAlternatingRowColors(True)
    table.setStyleSheet('alternate-background-color: #262E38;')
    table.horizontalHeader().setResizeMode(QHeaderView.Stretch)
    return update_table(table, data)



from pkg_resources import get_distribution
from PyQt5.QtWidgets import QMainWindow, QMessageBox, QStackedWidget, QDesktopWidget
# from PyQt5.QtWidgets import QMenuBar, QMenu
from ..pages.home_page import BadgerHomePage
from ..pages.routine_page import BadgerRoutinePage
from ..pages.run_page import BadgerRunPage


class BadgerMainWindow(QMainWindow):
    def __init__(self):
        super().__init__()

        self.init_ui()
        self.config_logic()

    def init_ui(self):
        version = get_distribution('badger-opt').version
        self.setWindowTitle(f'Badger v{version}')
        self.resize(960, 720)
        self.center()

        # Add menu bar
        # menu_bar = self.menuBar()
        # edit_menu = menu_bar.addMenu('Edit')
        # edit_menu.addAction('New')

        # Add pages
        self.home_page = BadgerHomePage(self.show_routine_page, self.show_run_page)
        self.routine_page = BadgerRoutinePage(self.show_home_page)
        self.run_page = BadgerRunPage(self.show_routine_page, self.show_home_page)

        self.stacks = stacks = QStackedWidget()
        stacks.addWidget(self.home_page)
        stacks.addWidget(self.routine_page)
        stacks.addWidget(self.run_page)

        stacks.setCurrentIndex(0)

        self.setCentralWidget(self.stacks)

    def center(self):
        qr = self.frameGeometry()
        cp = QDesktopWidget().availableGeometry().center()

        qr.moveCenter(cp)
        self.move(qr.topLeft())

    def config_logic(self):
        pass

    def show_routine_page(self, routine=None):
        try:
            self.routine_page.refresh_ui(routine)
        except Exception as e:
            return QMessageBox.critical(self, 'Error!', str(e))

        self.stacks.setCurrentIndex(1)

    def show_home_page(self):
        self.routine_page.refresh_ui(None)  # clean up to prevent bugs
        self.home_page.refresh_ui()
        self.home_page.reconfig_logic()
        self.stacks.setCurrentIndex(0)

    def show_run_page(self):
        try:
            self.run_page.refresh_ui()
            self.run_page.reconfig_logic()
        except Exception as e:
            return QMessageBox.critical(self, 'Error!', str(e))

        self.stacks.setCurrentIndex(2)



import logging
logger = logging.getLogger(__name__)
import time
import pandas as pd
from coolname import generate_slug
from ..utils import load_config, merge_params, normalize_routine
from ..utils import config_list_to_dict, curr_ts, ts_to_str
from ..utils import run_routine as run


def run_n_archive(routine, yes=False, save=False, verbose=2,
                  fmt='lcls-log-full', sleep=0):
    try:
        from ..archive import archive_run
    except Exception as e:
        logger.error(e)
        return

    var_names = [next(iter(d)) for d in routine['config']['variables']]
    obj_names = [next(iter(d)) for d in routine['config']['objectives']]
    if routine['config']['constraints']:
        con_names = [next(iter(d)) for d in routine['config']['constraints']]
    else:
        con_names = []
    # Make data a list to avoid global var def
    data = [pd.DataFrame(None, columns=['timestamp_raw', 'timestamp'] + obj_names + con_names + var_names)]

    def after_evaluate(vars, obses, cons):
        # vars: ndarray
        # obses: ndarray
        # cons: ndarray
        ts = curr_ts()
        solution = [ts.timestamp(), ts_to_str(ts, fmt)] + list(obses) + list(cons) + list(vars)
        data[0] = data[0].append(pd.Series(solution, index=data[0].columns), ignore_index=True)
        # take a break to let the outside signal to change the status
        time.sleep(sleep)

    try:
        run(routine, yes, save, verbose, after_evaluate=after_evaluate)
    except Exception as e:
        logger.error(e)

    archive_run(routine, data[0])


def run_routine(args):
    try:
        from ..factory import get_algo, get_env
    except Exception as e:
        logger.error(e)
        return

    try:
        # Get env params
        _, configs_env = get_env(args.env)

        # Get algo params
        _, configs_algo = get_algo(args.algo)

        # Normalize the algo and env params
        params_env = load_config(args.env_params)
        params_algo = load_config(args.algo_params)
    except Exception as e:
        logger.error(e)
        return
    params_env = merge_params(configs_env['params'], params_env)
    params_algo = merge_params(configs_algo['params'], params_algo)

    # Load routine configs
    try:
        configs_routine = load_config(args.config)
    except Exception as e:
        logger.error(e)
        return

    # Compose the routine
    routine = {
        'name': args.save or generate_slug(2),
        'algo': args.algo,
        'env': args.env,
        'algo_params': params_algo,
        'env_params': params_env,
        # env_vranges is an additional info for the normalization
        # Will be removed after the normalization
        'env_vranges': config_list_to_dict(configs_env['variables']),
        'config': configs_routine,
    }

    # Sanity check and config normalization
    try:
        routine = normalize_routine(routine)
    except Exception as e:
        logger.error(e)
        return

    run_n_archive(routine, args.yes, args.save, args.verbose)



import os
import subprocess
from pkg_resources import get_distribution


def capture(command):
    proc = subprocess.Popen(
        command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = proc.communicate()
    out = out.decode('utf-8')
    err = err.decode('utf-8')
    return out, err, proc.returncode


def test_cli_main():
    command = ['badger']
    out, err, exitcode = capture(command)

    assert exitcode == 0

    # Check output lines
    outlines = out.split('\n')
    assert len(outlines) == 7

    # Check name
    assert outlines[0] == 'name: Badger the optimizer'

    # Check version
    version = get_distribution('badger-opt').version
    assert outlines[1] == f'version: {version}'


def test_cli_run(mock_config_root):
    command = ['badger', 'run', '-a', 'silly', '-ap',
               '{dimension: 2, max_iter: 10}',  '-e', 'silly', '-c',
               os.path.join(mock_config_root, 'test.yaml'), '-y']
    out, err, exitcode = capture(command)

    assert exitcode == 0

    # Check output lines
    outlines = out.split('\n')
    assert len(outlines) == 15

    # Check table header
    assert outlines[1] == '|    iter    |     l2     |     q1     |     q2     |'



import numpy as np
from PyQt5.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout
from PyQt5.QtWidgets import QPushButton, QSplitter, QTextEdit, QShortcut
from PyQt5.QtWidgets import QListWidget, QListWidgetItem, QLabel
from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtGui import QKeySequence
from ..components.search_bar import search_bar
from ..components.data_table import data_table, update_table, reset_table, add_row
from ..components.routine_item import BadgerRoutineItem
from ..components.history_navigator import HistoryNavigator
from ..components.run_monitor import BadgerOptMonitor
from ..components.routine_editor import BadgerRoutineEditor
from ..components.status_bar import BadgerStatusBar
from ...db import list_routine, load_routine, remove_routine, get_runs_by_routine, get_runs
from ...archive import load_run, delete_run
from ...utils import get_header


stylesheet = '''
QPushButton:hover:pressed
{
    background-color: #C7737B;
}
QPushButton:hover
{
    background-color: #BF616A;
}
QPushButton
{
    background-color: #A9444E;
}
'''

class BadgerHomePage(QWidget):
    sig_routine_activated = pyqtSignal(bool)

    def __init__(self):
        super().__init__()

        self.mode = 'regular'  # home page mode
        self.splitter_state = None  # store the run splitter state

        self.init_ui()
        self.config_logic()

        self.load_all_runs()

    def init_ui(self):
        # Set up the layout
        vbox = QVBoxLayout(self)
        vbox.setContentsMargins(0, 0, 0, 0)
        splitter = QSplitter(Qt.Horizontal)
        splitter.setStretchFactor(0, 0)
        splitter.setStretchFactor(1, 1)
        # splitter.setSizes([100, 200])
        vbox.addWidget(splitter, 1)

        # Routine panel
        self.panel_routine = panel_routine = QWidget()
        vbox_routine = QVBoxLayout(panel_routine)

        # Search bar
        panel_search = QWidget()
        hbox_search = QHBoxLayout(panel_search)
        hbox_search.setContentsMargins(0, 0, 0, 0)

        self.sbar = sbar = search_bar()
        sbar.setFixedHeight(36)
        f = sbar.font()
        f.setPixelSize(16)
        sbar.setFont(f)
        self.btn_new = btn_new = QPushButton('+')
        btn_new.setFixedSize(36, 36)
        btn_new.setFont(f)
        hbox_search.addWidget(sbar)
        hbox_search.addWidget(btn_new)
        vbox_routine.addWidget(panel_search)

        # Routine list
        self.routine_list = routine_list = QListWidget()
        routine_list.setAlternatingRowColors(True)
        routine_list.setSpacing(1)
        routine_list.setViewportMargins(0, 0, 17, 0)  # leave space for scrollbar
        self.build_routine_list()
        self.prev_routine = None  # last selected routine
        vbox_routine.addWidget(routine_list)

        # Info panel
        panel_info = QWidget()
        vbox_info = QVBoxLayout(panel_info)

        panel_view = QWidget()
        vbox_view = QVBoxLayout(panel_view)
        vbox_view.setContentsMargins(0, 0, 0, 0)
        vbox_info.addWidget(panel_view)

        # History run nav
        self.history_nav_bar = history_nav_bar = QWidget()
        hbox_nav = QHBoxLayout(history_nav_bar)
        hbox_nav.setContentsMargins(0, 0, 0, 0)
        vbox_view.addWidget(history_nav_bar)

        label_nav = QLabel('History Run')
        self.cb_history = cb_history = HistoryNavigator()
        self.btn_prev = btn_prev = QPushButton('<')
        self.btn_next = btn_next = QPushButton('>')
        btn_prev.setFixedWidth(32)
        btn_next.setFixedWidth(32)
        btn_prev.setDisabled(True)
        btn_next.setDisabled(True)
        hbox_nav.addWidget(label_nav)
        hbox_nav.addWidget(cb_history, 1)
        hbox_nav.addWidget(btn_prev)
        hbox_nav.addWidget(btn_next)

        # Routine edit + run monitor + data table
        self.splitter_run = splitter_run = QSplitter(Qt.Vertical)
        splitter_run.setStretchFactor(0, 0)
        splitter_run.setStretchFactor(1, 1)
        splitter_run.setStretchFactor(2, 0)
        vbox_view.addWidget(splitter_run)

        # Run monitor
        self.run_view = run_view = QWidget()  # for consistent bg
        vbox_run_view = QVBoxLayout(run_view)
        vbox_run_view.setContentsMargins(0, 10, 0, 10)
        self.run_monitor = run_monitor = BadgerOptMonitor()
        vbox_run_view.addWidget(run_monitor)

        # Data table
        self.run_table = run_table = data_table()

        # Routine view
        self.routine_view = routine_view = QWidget()  # for consistent bg
        vbox_routine_view = QVBoxLayout(routine_view)
        vbox_routine_view.setContentsMargins(0, 0, 0, 10)
        self.routine_editor = routine_editor = BadgerRoutineEditor()
        vbox_routine_view.addWidget(routine_editor)

        splitter_run.addWidget(routine_view)
        splitter_run.addWidget(run_view)
        splitter_run.addWidget(run_table)
        splitter_run.setSizes([0, 1, 0])  # collapse routine/table by default

        # Add panels to splitter
        splitter.addWidget(panel_routine)
        splitter.addWidget(panel_info)

        self.status_bar = status_bar = BadgerStatusBar()
        status_bar.set_summary('Badger is ready!')
        vbox_view.addWidget(status_bar)

    def config_logic(self):
        self.colors = ['c', 'g', 'm', 'y', 'b', 'r', 'w']
        self.symbols = ['o', 't', 't1', 's', 'p', 'h', 'd']

        self.sbar.textChanged.connect(self.build_routine_list)
        self.btn_new.clicked.connect(self.create_new_routine)
        self.routine_list.itemClicked.connect(self.select_routine)
        self.run_table.cellClicked.connect(self.solution_selected)
        self.run_table.itemSelectionChanged.connect(self.table_selection_changed)

        self.cb_history.currentIndexChanged.connect(self.go_run)
        self.btn_prev.clicked.connect(self.go_prev_run)
        self.btn_next.clicked.connect(self.go_next_run)

        self.run_monitor.sig_inspect.connect(self.inspect_solution)
        self.run_monitor.sig_lock.connect(self.toggle_lock)
        self.run_monitor.sig_new_run.connect(self.new_run)
        self.run_monitor.sig_run_name.connect(self.run_name)
        self.run_monitor.sig_progress.connect(self.progress)
        self.run_monitor.sig_del.connect(self.delete_run)

        self.routine_editor.sig_saved.connect(self.routine_saved)
        self.routine_editor.sig_canceled.connect(self.done_create_routine)
        self.routine_editor.sig_deleted.connect(self.routine_deleted)

        # Assign shortcuts
        self.shortcut_go_search = QShortcut(QKeySequence('Ctrl+L'), self)
        self.shortcut_go_search.activated.connect(self.go_search)

    def go_search(self):
        self.sbar.setFocus()

    def load_all_runs(self):
        runs = get_runs()
        self.cb_history.updateItems(runs)

    def create_new_routine(self):
        self.splitter_state = self.splitter_run.saveState()
        self.routine_editor.set_routine(None)
        self.splitter_run.setSizes([1, 0, 0])
        self.mode = 'new routine'
        self.routine_editor.switch_mode(self.mode)
        self.toggle_lock(True)
        self.run_monitor.setDisabled(True)
        self.run_table.setDisabled(True)

    def select_routine(self, item):
        if self.prev_routine:
            try:
                self.routine_list.itemWidget(self.prev_routine).deactivate()
            except:
                pass

            if self.prev_routine.routine == item.routine:  # click a routine again to deselect
                self.prev_routine = None
                self.current_routine = None
                self.load_all_runs()
                if not self.cb_history.count():
                    self.go_run(-1)  # sometimes we need to trigger this manually
                self.sig_routine_activated.emit(False)
                return

        self.prev_routine = item  # note that prev_routine is an item!
        self.sig_routine_activated.emit(True)

        routine, timestamp = load_routine(item.routine)
        self.current_routine = routine
        self.routine_editor.set_routine(routine)
        runs = get_runs_by_routine(routine['name'])
        self.cb_history.updateItems(runs)
        if not self.cb_history.count():
            self.go_run(-1)  # sometimes we need to trigger this manually

        if not runs:  # auto plot will not be triggered
            self.run_monitor.init_plots(routine)

        self.routine_list.itemWidget(item).activate()

    def build_routine_list(self, keyword=''):
        routines, timestamps = list_routine(keyword)

        try:
            selected_routine = self.prev_routine.routine
        except:
            selected_routine = None
        self.routine_list.clear()
        for i, routine in enumerate(routines):
            _item = BadgerRoutineItem(routine, timestamps[i])
            _item.sig_del.connect(self.delete_routine)
            item = QListWidgetItem(self.routine_list)
            item.routine = routine  # dirty trick
            item.setSizeHint(_item.sizeHint())
            self.routine_list.addItem(item)
            self.routine_list.setItemWidget(item, _item)
            if routine == selected_routine:
                _item.activate()
                self.prev_routine = item

    def go_run(self, i):
        if self.cb_history.currentText() == 'Optimization in progress...':
            return

        self.btn_prev.setDisabled(self.cb_history.currentIsFirst())
        self.btn_next.setDisabled(self.cb_history.currentIsLast())

        if i == -1:
            update_table(self.run_table)
            self.run_monitor.init_plots(self.current_routine)
            if not self.current_routine:
                self.routine_editor.clear()
                self.status_bar.set_summary('no active routine')
            else:
                self.status_bar.set_summary(f'current routine: {self.current_routine["name"]}')
            return

        run_filename = self.cb_history.currentText()
        try:
            run = load_run(run_filename)
        except:
            return
        self.current_routine = run['routine']  # update the current routine
        update_table(self.run_table, run['data'])
        self.run_monitor.init_plots(run['routine'], run['data'], run_filename)
        self.routine_editor.set_routine(run['routine'])
        self.status_bar.set_summary(f'current routine: {self.current_routine["name"]}')

    def go_prev_run(self):
        self.cb_history.selectPreviousItem()

    def go_next_run(self):
        self.cb_history.selectNextItem()

    def inspect_solution(self, idx):
        self.run_table.selectRow(idx)

    def solution_selected(self, r, c):
        self.run_monitor.jump_to_solution(r)

    def table_selection_changed(self):
        indices = self.run_table.selectedIndexes()
        if len(indices) == 1:  # let other method handles it
            return

        row = -1
        for index in indices:
            _row = index.row()
            if _row == row:
                continue

            if row == -1:
                row = _row
                continue

            return

        self.run_monitor.jump_to_solution(row)

    def toggle_lock(self, lock):
        if lock:
            self.panel_routine.setDisabled(True)
            self.history_nav_bar.setDisabled(True)
        else:
            self.panel_routine.setDisabled(False)
            self.history_nav_bar.setDisabled(False)

    def new_run(self):
        self.cb_history.insertItem(0, 'Optimization in progress...')
        self.cb_history.setCurrentIndex(0)

        header = get_header(self.current_routine)
        reset_table(self.run_table, header)

    def run_name(self, name):
        if self.prev_routine:
            runs = get_runs_by_routine(self.current_routine['name'])
        else:
            runs = get_runs()
        self.cb_history.updateItems(runs)

    def progress(self, vars, objs, cons):
        add_row(self.run_table, objs + cons + vars)

    def delete_run(self):
        run_name = self.cb_history.currentText()
        delete_run(run_name)
        # Reset current routine if no routine is selected
        if not self.prev_routine:
            self.current_routine = None
            runs = get_runs()
        else:
            runs = get_runs_by_routine(self.current_routine['name'])
        self.cb_history.updateItems(runs)
        if not self.cb_history.count():
            self.go_run(-1)  # sometimes we need to trigger this manually

    def routine_saved(self):
        keyword = self.sbar.text()
        self.build_routine_list(keyword)
        self.done_create_routine()

    def done_create_routine(self):
        if self.mode == 'new routine':
            self.mode = 'regular'
            self.routine_editor.switch_mode(self.mode)
            self.splitter_run.restoreState(self.splitter_state)
            self.splitter_state = None
            self.toggle_lock(False)
            self.run_monitor.setDisabled(False)
            self.run_table.setDisabled(False)

    def delete_routine(self, name):
        remove_routine(name)
        self.routine_deleted(name)

    def routine_deleted(self, name=None):
        if self.prev_routine:
            if self.prev_routine.routine == name:
                self.prev_routine = None
                self.current_routine = None
                self.load_all_runs()
                if not self.cb_history.count():
                    self.go_run(-1)  # sometimes we need to trigger this manually
                self.sig_routine_activated.emit(False)

        keyword = self.sbar.text()
        self.build_routine_list(keyword)



import numpy as np
import scipy.optimize as sopt
from operator import itemgetter
import logging


def optimize(evaluate, params):
    D, x0, isteps, xtol, max_iter = \
        itemgetter('dimension', 'x0', 'isteps', 'xtol', 'max_iter')(params)

    assert len(x0) == D, 'Dimension does not match!'

    if isteps is None or len(isteps) != D:
        logging.warning("Initial simplex is None")
        isim = None
    elif np.count_nonzero(isteps) != D:
        logging.warning("There is zero step. Initial simplex is None")
        isim = None
    else:
        isim = np.zeros((D + 1, D))
        isim[0] = x0
        for i in range(D):
            vertex = np.zeros(D)
            vertex[i] = isteps[i]
            isim[i + 1] = x0 + vertex

    logging.debug(f'ISIM = {isim}')

    def _evaluate(x):
        y, _, _, _ = evaluate(np.array(x).reshape(1, -1))
        y = y[0]

        return y

    res = sopt.fmin(_evaluate, x0, maxiter=max_iter,
                    maxfun=max_iter, xtol=xtol, initial_simplex=isim)

    return res



import numpy as np
from badger import environment
from badger.interface import Interface


class Environment(environment.Environment):

    name = 'silly'
    var_channel_map = {
        'q1': 'c1',
        'q2': 'c2',
        'q3': 'c3',
        'q4': 'c4',
    }

    def __init__(self, interface: Interface, params):
        super().__init__(interface, params)

    @staticmethod
    def list_vars():
        return ['q1', 'q2', 'q3', 'q4']

    @staticmethod
    def list_obses():
        return ['l1', 'l2']

    @staticmethod
    def get_default_params():
        return None

    def _get_var(self, var):
        return self.interface.get_value(self.var_channel_map[var])

    def _set_var(self, var, x):
        self.interface.set_value(self.var_channel_map[var], x)

    def _get_obs(self, obs):
        if obs == 'l1':
            values = self.interface.get_values(
                ['c1', 'c2', 'c3', 'c4', 'c5', 'c6'])
            return np.sum(np.abs(values))
        elif obs == 'l2':
            return self.interface.get_value('norm')



from PyQt5.QtWidgets import QVBoxLayout, QHBoxLayout, QPushButton, QWidget, QPlainTextEdit
from PyQt5.QtWidgets import QComboBox, QCheckBox, QStyledItemDelegate, QLabel
from .collapsible_box import CollapsibleBox


class BadgerEnvBox(CollapsibleBox):
    def __init__(self, envs=[], parent=None):
        super().__init__(' Environment', parent)

        self.envs = envs

        self.init_ui()

    def init_ui(self):
        vbox = QVBoxLayout()

        name = QWidget()
        hbox_name = QHBoxLayout(name)
        hbox_name.setContentsMargins(0, 0, 0, 0)
        lbl = QLabel('Name')
        lbl.setFixedWidth(64)
        self.cb = cb = QComboBox()
        cb.setItemDelegate(QStyledItemDelegate())
        cb.addItems(self.envs)
        cb.setCurrentIndex(-1)
        hbox_name.addWidget(lbl)
        hbox_name.addWidget(cb, 1)
        vbox.addWidget(name)

        params = QWidget()
        hbox_params = QHBoxLayout(params)
        hbox_params.setContentsMargins(0, 0, 0, 0)
        lbl_params_col = QWidget()
        vbox_lbl_params = QVBoxLayout(lbl_params_col)
        vbox_lbl_params.setContentsMargins(0, 0, 0, 0)
        lbl_params = QLabel('Params')
        lbl_params.setFixedWidth(64)
        vbox_lbl_params.addWidget(lbl_params)
        vbox_lbl_params.addStretch(1)
        hbox_params.addWidget(lbl_params_col)

        edit_params_col = QWidget()
        vbox_params_edit = QVBoxLayout(edit_params_col)
        vbox_params_edit.setContentsMargins(0, 0, 0, 8)
        self.edit = edit = QPlainTextEdit()
        edit.setFixedHeight(128)
        vbox_params_edit.addWidget(edit)
        hbox_params.addWidget(edit_params_col)
        vbox.addWidget(params)

        action_bar = QWidget()
        hbox_action = QHBoxLayout(action_bar)
        hbox_action.setContentsMargins(0, 0, 0, 0)
        self.btn_env_play = btn_env_play = QPushButton('Open Playground')
        btn_env_play.setFixedSize(128, 24)
        # btn_env_play.hide()  # hide for now to prevent confusions
        hbox_action.addStretch(1)
        hbox_action.addWidget(btn_env_play)
        vbox.addWidget(action_bar)

        self.setContentLayout(vbox)



import os
from datetime import datetime
import logging
logger = logging.getLogger(__name__)
import yaml
import sqlite3
from .settings import read_value


# Check badger database root
BADGER_DB_ROOT = read_value('BADGER_DB_ROOT')
if BADGER_DB_ROOT is None:
    raise Exception('Please set the BADGER_DB_ROOT env var!')
elif not os.path.exists(BADGER_DB_ROOT):
    os.makedirs(BADGER_DB_ROOT)
    logger.info(
        f'Badger database root {BADGER_DB_ROOT} created')


def save_routine(routine):
    db_routine = os.path.join(BADGER_DB_ROOT, 'routines.db')

    # Initialize
    if not os.path.exists(db_routine):
        con = sqlite3.connect(db_routine)
        cur = con.cursor()
        cur.execute('create table routine (name not null primary key, config, savedAt timestamp)')
    else:
        con = sqlite3.connect(db_routine)
        cur = con.cursor()

    # Insert a record
    try:
        cur.execute('insert into routine values (?, ?, ?)',
                    (routine['name'], yaml.dump(routine, sort_keys=False), datetime.now()))
    except sqlite3.OperationalError:
        cur.execute('create table routine (name not null primary key, config, savedAt timestamp)')
        cur.execute('insert into routine values (?, ?, ?)',
                    (routine['name'], yaml.dump(routine, sort_keys=False), datetime.now()))
    con.commit()
    con.close()


def remove_routine(name, remove_runs=False):
    db_routine = os.path.join(BADGER_DB_ROOT, 'routines.db')

    con = sqlite3.connect(db_routine)
    cur = con.cursor()

    cur.execute(f'delete from routine where name = "{name}"')

    con.commit()
    con.close()

    if remove_runs:
        # Remove all related run records
        db_run = os.path.join(BADGER_DB_ROOT, 'runs.db')

        con = sqlite3.connect(db_run)
        cur = con.cursor()

        cur.execute(f'delete from run where routine = "{name}"')

        con.commit()
        con.close()


def load_routine(name):
    db_routine = os.path.join(BADGER_DB_ROOT, 'routines.db')
    con = sqlite3.connect(db_routine)
    cur = con.cursor()

    try:
        cur.execute('select * from routine where name=:name', {'name': name})
    except sqlite3.OperationalError:
        cur.execute('create table routine (name not null primary key, config, savedAt timestamp)')
        cur.execute('select * from routine where name=:name', {'name': name})
    records = cur.fetchall()
    con.close()

    if len(records) == 1:
        return yaml.safe_load(records[0][1]), records[0][2]
    elif len(records) == 0:
        # logger.warning(f'Routine {name} not found in the database!')
        return None, None
    else:
        raise Exception(
            f'Multiple routines with name {name} found in the database!')


def list_routine(keyword=''):
    db_routine = os.path.join(BADGER_DB_ROOT, 'routines.db')
    con = sqlite3.connect(db_routine)
    cur = con.cursor()

    try:
        cur.execute(f'select name, savedAt from routine where name like "%{keyword}%" order by savedAt desc')
    except sqlite3.OperationalError:
        cur.execute('create table routine (name not null primary key, config, savedAt timestamp)')
        cur.execute(f'select name, savedAt from routine where name like "%{keyword}%" order by savedAt desc')
    records = cur.fetchall()
    names = [record[0] for record in records]
    timestamps = [record[1] for record in records]
    con.close()

    return names, timestamps


def save_run(run):
    db_run = os.path.join(BADGER_DB_ROOT, 'runs.db')

    # Initialize
    if not os.path.exists(db_run):
        con = sqlite3.connect(db_run)
        cur = con.cursor()
        cur.execute('create table run (id integer primary key, savedAt timestamp, finishedAt timestamp, routine, filename)')
    else:
        con = sqlite3.connect(db_run)
        cur = con.cursor()

    # Insert a record
    routine_name = run['routine']['name']
    run_filename = run['filename']
    timestamps = run['data']['timestamp_raw']
    time_start = datetime.fromtimestamp(timestamps[0])
    time_finish = datetime.fromtimestamp(timestamps[-1])
    try:
        cur.execute('insert into run values (?, ?, ?, ?, ?)',
                    (None, time_start, time_finish, routine_name, run_filename))
    except sqlite3.OperationalError:
        cur.execute('create table run (id integer primary key, savedAt timestamp, finishedAt timestamp, routine, filename)')
        cur.execute('insert into run values (?, ?, ?, ?, ?)',
                    (None, time_start, time_finish, routine_name, run_filename))
    rid = cur.lastrowid

    con.commit()
    con.close()

    return rid


def get_runs_by_routine(routine: str):
    db_run = os.path.join(BADGER_DB_ROOT, 'runs.db')

    con = sqlite3.connect(db_run)
    cur = con.cursor()

    cur.execute(f'select filename from run where routine = "{routine}" order by savedAt desc')
    records = cur.fetchall()
    con.close()

    filenames = [record[0] for record in records]

    return filenames


def get_runs():
    db_run = os.path.join(BADGER_DB_ROOT, 'runs.db')

    con = sqlite3.connect(db_run)
    cur = con.cursor()

    cur.execute(f'select filename from run order by savedAt desc')
    records = cur.fetchall()
    con.close()

    filenames = [record[0] for record in records]

    return filenames


def remove_run_by_filename(name):
    db_run = os.path.join(BADGER_DB_ROOT, 'runs.db')

    con = sqlite3.connect(db_run)
    cur = con.cursor()

    cur.execute(f'delete from run where filename = "{name}"')

    con.commit()
    con.close()


def remove_run_by_id(rid):
    db_run = os.path.join(BADGER_DB_ROOT, 'runs.db')

    con = sqlite3.connect(db_run)
    cur = con.cursor()

    cur.execute(f'delete from run where id = {rid}')

    con.commit()
    con.close()



import numpy as np
from badger import interface
from operator import itemgetter
import logging


class Interface(interface.Interface):

    name = 'silly'

    def __init__(self, params=None):
        super().__init__(params)

        prefix, count = itemgetter(
            'channel_prefix', 'channel_count')(self.params)

        self.channels = []
        self.states = {}
        for i in range(count):
            self.channels.append(f'{prefix}{i + 1}')
            self.states[f'{prefix}{i + 1}'] = 0

        self.channels.append('norm')
        self.states['norm'] = 0

    @staticmethod
    def get_default_params():
        return {
            'channel_prefix': 'c',
            'channel_count': 8
        }

    def get_value(self, channel: str):
        try:
            value = self.states[channel]
        except KeyError:
            logging.warning(f'Channel {channel} doesn\'t exist!')
            value = None

        return value

    def set_value(self, channel: str, value):
        if channel not in self.channels:
            logging.warning(f'Channel {channel} doesn\'t exist!')
            return

        try:
            self.states[channel] = value
            values = np.array([self.states[channel]
                              for channel in self.channels[:-1]])
            self.states['norm'] = np.sqrt(np.sum(values ** 2))
        except KeyError:
            logging.warning(f'Channel {channel} doesn\'t exist!')



from PyQt5.QtWidgets import QVBoxLayout, QHBoxLayout, QWidget, QPushButton
from PyQt5.QtWidgets import QTextEdit, QLabel
from PyQt5.QtWidgets import QSizePolicy
from PyQt5.QtGui import QFont
from PyQt5.QtCore import Qt


class BadgerStatusBar(QWidget):
    def __init__(self):
        super().__init__()

        self.init_ui()
        self.config_logic()

    def init_ui(self):
        vbox = QVBoxLayout(self)
        vbox.setContentsMargins(0, 0, 0, 0)

        self.summary = summary = QLabel()
        summary.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        summary.setAlignment(Qt.AlignCenter)
        vbox.addWidget(summary)

    def config_logic(self):
        pass

    def set_summary(self, text):
        self.summary.setText(text)



from PyQt5.QtWidgets import QVBoxLayout, QHBoxLayout, QPushButton, QWidget
from PyQt5.QtWidgets import QComboBox, QCheckBox, QListWidget, QLabel
from .collapsible_box import CollapsibleBox


class BadgerConfigBox(CollapsibleBox):
    def __init__(self, parent=None):
        super().__init__(' Configs', parent)

        self.init_ui()
        self.config_logic()

    def init_ui(self):
        self.vbox = vbox = QVBoxLayout()

        # Variables config
        var_panel = QWidget()
        vbox.addWidget(var_panel, 2)
        hbox_var = QHBoxLayout(var_panel)
        hbox_var.setContentsMargins(0, 0, 0, 0)
        lbl_var_col = QWidget()
        vbox_lbl_var = QVBoxLayout(lbl_var_col)
        vbox_lbl_var.setContentsMargins(0, 0, 0, 0)
        lbl_var = QLabel('Vars')
        lbl_var.setFixedWidth(64)
        vbox_lbl_var.addWidget(lbl_var)
        vbox_lbl_var.addStretch(1)
        hbox_var.addWidget(lbl_var_col)

        edit_var_col = QWidget()
        vbox_var_edit = QVBoxLayout(edit_var_col)
        vbox_var_edit.setContentsMargins(0, 0, 0, 0)
        action_var = QWidget()
        hbox_action_var = QHBoxLayout(action_var)
        hbox_action_var.setContentsMargins(0, 0, 0, 0)
        vbox_var_edit.addWidget(action_var)
        self.btn_all_var = btn_all_var = QPushButton('Check All')
        self.btn_un_all_var = btn_un_all_var = QPushButton('Uncheck All')
        btn_all_var.setFixedSize(96, 24)
        btn_un_all_var.setFixedSize(96, 24)
        self.btn_add_var = btn_add_var = QPushButton('Add')
        btn_add_var.setFixedSize(96, 24)
        btn_add_var.setDisabled(True)
        self.check_only_var = check_only_var = QCheckBox('Show Checked Only')
        check_only_var.setChecked(False)
        hbox_action_var.addWidget(btn_all_var)
        hbox_action_var.addWidget(btn_un_all_var)
        hbox_action_var.addWidget(btn_add_var)
        hbox_action_var.addStretch()
        hbox_action_var.addWidget(check_only_var)
        self.list_var = QListWidget()
        self.list_var.setViewportMargins(2, 2, 17, 2)
        # self.list_var.setFixedHeight(128)
        vbox_var_edit.addWidget(self.list_var)
        # vbox_var_edit.addStretch()
        hbox_var.addWidget(edit_var_col)

        # Objectives config
        obj_panel = QWidget()
        vbox.addWidget(obj_panel, 1)
        hbox_obj = QHBoxLayout(obj_panel)
        hbox_obj.setContentsMargins(0, 0, 0, 0)
        lbl_obj_col = QWidget()
        vbox_lbl_obj = QVBoxLayout(lbl_obj_col)
        vbox_lbl_obj.setContentsMargins(0, 0, 0, 0)
        lbl_obj = QLabel('Objs')
        lbl_obj.setFixedWidth(64)
        vbox_lbl_obj.addWidget(lbl_obj)
        vbox_lbl_obj.addStretch(1)
        hbox_obj.addWidget(lbl_obj_col)

        edit_obj_col = QWidget()
        vbox_obj_edit = QVBoxLayout(edit_obj_col)
        vbox_obj_edit.setContentsMargins(0, 0, 0, 0)
        action_obj = QWidget()
        hbox_action_obj = QHBoxLayout(action_obj)
        hbox_action_obj.setContentsMargins(0, 0, 0, 0)
        vbox_obj_edit.addWidget(action_obj)
        self.btn_all_obj = btn_all_obj = QPushButton('Check All')
        self.btn_un_all_obj = btn_un_all_obj = QPushButton('Uncheck All')
        btn_all_obj.setFixedSize(96, 24)
        btn_un_all_obj.setFixedSize(96, 24)
        self.check_only_obj = check_only_obj = QCheckBox('Show Checked Only')
        check_only_obj.setChecked(False)
        hbox_action_obj.addWidget(btn_all_obj)
        hbox_action_obj.addWidget(btn_un_all_obj)
        hbox_action_obj.addStretch()
        hbox_action_obj.addWidget(check_only_obj)
        self.list_obj = QListWidget()
        # self.list_obj.setFixedHeight(64)
        self.list_obj.setViewportMargins(2, 2, 17, 2)
        vbox_obj_edit.addWidget(self.list_obj)
        # vbox_obj_edit.addStretch()
        hbox_obj.addWidget(edit_obj_col)

        # Constraints config
        con_panel = QWidget()
        vbox.addWidget(con_panel, 1)
        hbox_con = QHBoxLayout(con_panel)
        hbox_con.setContentsMargins(0, 0, 0, 0)
        lbl_con_col = QWidget()
        vbox_lbl_con = QVBoxLayout(lbl_con_col)
        vbox_lbl_con.setContentsMargins(0, 0, 0, 0)
        lbl_con = QLabel('Cons')
        lbl_con.setFixedWidth(64)
        vbox_lbl_con.addWidget(lbl_con)
        vbox_lbl_con.addStretch(1)
        hbox_con.addWidget(lbl_con_col)

        edit_con_col = QWidget()
        vbox_con_edit = QVBoxLayout(edit_con_col)
        vbox_con_edit.setContentsMargins(0, 0, 0, 0)
        action_con = QWidget()
        hbox_action_con = QHBoxLayout(action_con)
        hbox_action_con.setContentsMargins(0, 0, 0, 0)
        vbox_con_edit.addWidget(action_con)
        self.btn_add_con = btn_add_con = QPushButton('Add')
        btn_add_con.setFixedSize(96, 24)
        btn_add_con.setDisabled(True)
        hbox_action_con.addWidget(btn_add_con)
        hbox_action_con.addStretch()
        self.list_con = QListWidget()
        # self.list_con.setFixedHeight(64)
        self.list_con.setViewportMargins(2, 2, 17, 2)
        vbox_con_edit.addWidget(self.list_con)
        # vbox_con_edit.addStretch()
        hbox_con.addWidget(edit_con_col)

        self.setContentLayout(vbox)

    def config_logic(self):
        self.dict_var = {}
        self.dict_obj = {}
        self.dict_con = {}

    def _fit_content(self, list):
        height = list.sizeHintForRow(0) * list.count() + 2 * list.frameWidth() + 4
        height = max(28, min(height, 192))
        list.setFixedHeight(height)

    def fit_content(self):
        return
        self._fit_content(self.list_var)
        self._fit_content(self.list_obj)
        self._fit_content(self.list_con)



from datetime import datetime
from PyQt5.QtWidgets import QWidget, QHBoxLayout, QVBoxLayout, QLabel, QPushButton, QSizePolicy
from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtGui import QFont


stylesheet_normal = '''
    background-color: #4C566A;
    border-radius: 2px;
'''

stylesheet_normal_hover = '''
    background-color: #5E81AC;
    border-radius: 2px;
'''

stylesheet_activate = '''
    background-color: #4B6789;
    border-radius: 2px;
'''

stylesheet_activate_hover = '''
    background-color: #54749A;
    border-radius: 2px;
'''

stylesheet_del = '''
QPushButton:hover:pressed
{
    background-color: #BF616A;
}
QPushButton:hover
{
    background-color: #A9444E;
}
QPushButton
{
    background-color: none;
    border-top-left-radius: 0px;
    border-bottom-left-radius: 0px;
}
'''

class BadgerRoutineItem(QWidget):
    sig_del = pyqtSignal(str)

    def __init__(self, name, timestamp, parent=None):
        super().__init__(parent)

        self.activated = False
        self.hover = False
        self.name = name
        self.timestamp = timestamp

        self.init_ui()
        self.config_logic()

    def init_ui(self):
        self.setAttribute(Qt.WA_StyledBackground)
        self.setStyleSheet(stylesheet_normal)

        cool_font = QFont()
        cool_font.setWeight(QFont.DemiBold)
        cool_font.setPixelSize(16)

        hbox = QHBoxLayout(self)
        hbox.setContentsMargins(0, 0, 0, 0)
        hbox.setSpacing(0)
        info_panel = QWidget()
        hbox.addWidget(info_panel, 1)
        vbox = QVBoxLayout(info_panel)
        vbox.setContentsMargins(8, 8, 8, 8)
        vbox.setSpacing(0)
        routine_name = QLabel(self.name)
        routine_name.setFont(cool_font)
        vbox.addWidget(routine_name)
        _timestamp = datetime.fromisoformat(self.timestamp)
        time_str = _timestamp.strftime('%m/%d/%Y, %H:%M:%S')
        time_created = QLabel(time_str)
        vbox.addWidget(time_created)

        self.btn_del = btn_del = QPushButton('X')
        btn_del.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)
        btn_del.setFixedWidth(24)
        btn_del.setStyleSheet(stylesheet_del)
        btn_del.hide()
        hbox.addWidget(btn_del)

    def config_logic(self):
        self.btn_del.clicked.connect(self.delete_routine)

    def activate(self):
        self.activated = True
        if self.hover:
            self.setStyleSheet(stylesheet_activate_hover)
        else:
            self.setStyleSheet(stylesheet_activate)

    def deactivate(self):
        self.activated = False
        if self.hover:
            self.setStyleSheet(stylesheet_normal_hover)
        else:
            self.setStyleSheet(stylesheet_normal)

    def enterEvent(self, event):
        self.hover = True
        self.btn_del.show()
        if self.activated:
            self.setStyleSheet(stylesheet_activate_hover)
        else:
            self.setStyleSheet(stylesheet_normal_hover)

    def leaveEvent(self, event):
        self.hover = False
        self.btn_del.hide()
        if self.activated:
            self.setStyleSheet(stylesheet_activate)
        else:
            self.setStyleSheet(stylesheet_normal)

    def delete_routine(self):
        self.sig_del.emit(self.name)



import logging
logger = logging.getLogger(__name__)
from ..utils import range_to_str, yprint, run_routine


def show_routine(args):
    try:
        from ..db import load_routine, list_routine
        from .run import run_n_archive
    except Exception as e:
        logger.error(e)
        return

    # List routines
    if args.routine_name is None:
        routines = list_routine()[0]
        if routines:
            yprint(routines)
        else:
            print('No routine has been saved yet')
        return

    try:
        routine, _ = load_routine(args.routine_name)
        if routine is None:
            print(f'Routine {args.routine_name} not found')
            return
    except Exception as e:
        print(e)
        return

    # Print the routine
    if not args.run:
        routine['config']['variables'] = range_to_str(routine['config']['variables'])
        yprint(routine)
        return

    run_n_archive(routine, args.yes, False, args.verbose)



import time
import pandas as pd
from PyQt5.QtCore import pyqtSignal, QObject, QRunnable
from ...utils import curr_ts, run_routine, ts_to_str


class BadgerRoutineSignals(QObject):
    env_ready = pyqtSignal(list)
    finished = pyqtSignal()
    progress = pyqtSignal(list, list, list)
    error = pyqtSignal(Exception)
    info = pyqtSignal(str)


class BadgerRoutineRunner(QRunnable):

    def __init__(self, routine, save, verbose=2, use_full_ts=False):
        super().__init__()

        # Signals should belong to instance rather than class
        # Since there could be multiple runners runing in parallel
        self.signals = BadgerRoutineSignals()

        self.routine = routine
        self.run_filename = None
        self.var_names = var_names = [next(iter(d)) for d in routine['config']['variables']]
        self.obj_names = obj_names = [next(iter(d)) for d in routine['config']['objectives']]
        if routine['config']['constraints']:
            self.con_names = con_names = [next(iter(d)) for d in routine['config']['constraints']]
        else:
            self.con_names = con_names = []
        self.data = pd.DataFrame(None, columns=['timestamp_raw', 'timestamp'] + obj_names + con_names + var_names)
        self.save = save
        self.verbose = verbose
        self.use_full_ts = use_full_ts

        self.is_paused = False
        self.is_killed = False

    def run(self):
        error = None
        try:
            run_routine(self.routine, True, self.save, self.verbose,
                        self.before_evaluate, self.after_evaluate,
                        self.env_ready, self.pf_ready)
        except Exception as e:
            error = e

        self.signals.finished.emit()
        if error:
            if str(error) == 'Optimization run has been terminated!':
                self.signals.info.emit(str(error))
                return

            self.signals.error.emit(error)

    def before_evaluate(self, vars):
        # vars: ndarray
        while self.is_paused:
            time.sleep(0)
            if self.is_killed:
                raise Exception('Optimization run has been terminated!')

        if self.is_killed:
            raise Exception('Optimization run has been terminated!')

    def after_evaluate(self, vars, obses, cons):
        # vars: ndarray
        # obses: ndarray
        # cons: ndarray
        self.signals.progress.emit(list(vars), list(obses), list(cons))

        # Append solution to data
        fmt = 'lcls-log-full' if self.use_full_ts else 'lcls-log'
        ts = curr_ts()
        solution = [ts.timestamp(), ts_to_str(ts, fmt)] + list(obses) + list(cons) + list(vars)
        self.data = self.data.append(pd.Series(solution, index=self.data.columns), ignore_index=True)

        # take a break to let the outside signal to change the status
        time.sleep(0.1)

    def env_ready(self, env):
        self.env = env
        init_vars = env._get_vars(self.var_names)
        self.signals.env_ready.emit(init_vars)

    def pf_ready(self, pf):
        self.pf = pf

    def ctrl_routine(self, pause):
        self.is_paused = pause

    def stop_routine(self):
        self.is_killed = True



from PyQt5.QtWidgets import QLineEdit, QListWidget, QListWidgetItem, QWidget, QVBoxLayout, QHBoxLayout
from PyQt5.QtWidgets import QPushButton, QGroupBox, QComboBox, QLineEdit, QPlainTextEdit, QCheckBox
from PyQt5.QtWidgets import QMessageBox, QStyledItemDelegate
from PyQt5.QtCore import QSize
from PyQt5.QtGui import QFont
from coolname import generate_slug
from ...factory import list_algo, list_env, get_algo, get_env
from ...utils import ystring, load_config, config_list_to_dict, normalize_routine, instantiate_env
from ..components.variable_item import variable_item
from ..components.objective_item import objective_item
from ..components.constraint_item import constraint_item
from ..windows.review_dialog import BadgerReviewDialog
from ..windows.opt_monitor import BadgerOptMonitor
from ..windows.var_dialog import BadgerVariableDialog
from ..windows.edit_script_dialog import BadgerEditScriptDialog


CONS_RELATION_DICT = {
    '>': 'GREATER_THAN',
    '<': 'LESS_THAN',
    '=': 'EQUAL_TO',
}


class BadgerRoutinePage(QWidget):
    def __init__(self, go_home=None):
        super().__init__()

        self.go_home = go_home

        self.algos = list_algo()
        self.envs = list_env()
        self.env = None
        self.script = ''

        self.init_ui()
        self.config_logic()

    def init_ui(self):
        # Set up the layout
        vbox = QVBoxLayout(self)

        # Algo and env configs
        panel_int = QWidget()
        hbox_int = QHBoxLayout(panel_int)
        hbox_int.setContentsMargins(0, 0, 0, 0)

        group_algo = QGroupBox('Algorithm')
        vbox_algo = QVBoxLayout(group_algo)
        action_algo = QWidget()
        hbox_action_algo = QHBoxLayout(action_algo)
        hbox_action_algo.setContentsMargins(0, 0, 0, 0)
        self.cb_algo = cb_algo = QComboBox()
        cb_algo.setItemDelegate(QStyledItemDelegate())
        cb_algo.addItems(self.algos)
        cb_algo.setCurrentIndex(-1)
        self.check_use_script = check_use_script = QCheckBox('Use Script')
        check_use_script.setChecked(False)
        self.btn_edit_script = btn_edit_script = QPushButton('Edit Script')
        btn_edit_script.setFixedSize(96, 24)
        btn_edit_script.hide()
        hbox_action_algo.addWidget(cb_algo, 1)
        hbox_action_algo.addWidget(check_use_script)
        hbox_action_algo.addWidget(btn_edit_script)
        vbox_algo.addWidget(action_algo)
        self.edit_algo = edit_algo = QPlainTextEdit()
        edit_algo.setFixedHeight(128)
        vbox_algo.addWidget(edit_algo)
        hbox_int.addWidget(group_algo)

        group_env = QGroupBox('Environment')
        vbox_env = QVBoxLayout(group_env)
        action_env = QWidget()
        hbox_action_env = QHBoxLayout(action_env)
        hbox_action_env.setContentsMargins(0, 0, 0, 0)
        self.cb_env = cb_env = QComboBox()
        cb_env.setItemDelegate(QStyledItemDelegate())
        cb_env.addItems(self.envs)
        cb_env.setCurrentIndex(-1)
        self.btn_env_play = btn_env_play = QPushButton('Playground')
        btn_env_play.setFixedSize(96, 24)
        btn_env_play.hide()  # hide for now to prevent confusions
        hbox_action_env.addWidget(cb_env, 1)
        hbox_action_env.addWidget(btn_env_play)
        vbox_env.addWidget(action_env)
        self.edit_env = edit_env = QPlainTextEdit()
        edit_env.setFixedHeight(128)
        vbox_env.addWidget(edit_env)
        hbox_int.addWidget(group_env)
        vbox.addWidget(panel_int)

        # Configs group
        group_ext = QGroupBox('Configs')
        hbox_ext = QHBoxLayout(group_ext)

        group_var = QGroupBox('Variables')
        vbox_var = QVBoxLayout(group_var)
        action_var = QWidget()
        hbox_action_var = QHBoxLayout(action_var)
        hbox_action_var.setContentsMargins(0, 0, 0, 0)
        self.btn_all_var = btn_all_var = QPushButton('Check All')
        self.btn_un_all_var = btn_un_all_var = QPushButton('Uncheck All')
        btn_all_var.setFixedSize(96, 24)
        btn_un_all_var.setFixedSize(96, 24)
        self.btn_add_var = btn_add_var = QPushButton('Add')
        btn_add_var.setFixedSize(96, 24)
        btn_add_var.setDisabled(True)
        self.check_only_var = check_only_var = QCheckBox('Checked Only')
        check_only_var.setChecked(False)
        hbox_action_var.addWidget(btn_all_var)
        hbox_action_var.addWidget(btn_un_all_var)
        hbox_action_var.addWidget(btn_add_var)
        hbox_action_var.addStretch()
        hbox_action_var.addWidget(check_only_var)
        vbox_var.addWidget(action_var)
        self.list_var = list_var = QListWidget()
        self.dict_var = {}
        vbox_var.addWidget(list_var)
        hbox_ext.addWidget(group_var)

        group_obj = QGroupBox('Objectives')
        vbox_obj = QVBoxLayout(group_obj)
        action_obj = QWidget()
        hbox_action_obj = QHBoxLayout(action_obj)
        hbox_action_obj.setContentsMargins(0, 0, 0, 0)
        self.btn_all_obj = btn_all_obj = QPushButton('Check All')
        self.btn_un_all_obj = btn_un_all_obj = QPushButton('Uncheck All')
        btn_all_obj.setFixedSize(96, 24)
        btn_un_all_obj.setFixedSize(96, 24)
        self.check_only_obj = check_only_obj = QCheckBox('Checked Only')
        check_only_obj.setChecked(False)
        hbox_action_obj.addWidget(btn_all_obj)
        hbox_action_obj.addWidget(btn_un_all_obj)
        hbox_action_obj.addStretch()
        hbox_action_obj.addWidget(check_only_obj)
        vbox_obj.addWidget(action_obj)
        self.list_obj = list_obj = QListWidget()
        self.dict_obj = {}
        vbox_obj.addWidget(list_obj)
        hbox_ext.addWidget(group_obj)

        group_con = QGroupBox('Constraints')
        vbox_con = QVBoxLayout(group_con)
        action_con = QWidget()
        hbox_action_con = QHBoxLayout(action_con)
        hbox_action_con.setContentsMargins(0, 0, 0, 0)
        self.btn_add_con = btn_add_con = QPushButton('Add')
        btn_add_con.setFixedSize(96, 24)
        btn_add_con.setDisabled(True)
        hbox_action_con.addWidget(btn_add_con)
        hbox_action_con.addStretch()
        vbox_con.addWidget(action_con)
        self.list_con = list_con = QListWidget()
        self.dict_con = {}
        vbox_con.addWidget(list_con)
        hbox_ext.addWidget(group_con)

        vbox.addWidget(group_ext, 1)

        # Misc group
        group_misc = QGroupBox('Misc')
        hbox_misc = QHBoxLayout(group_misc)
        self.check_save = check_save = QCheckBox('Save as')
        check_save.setChecked(False)
        self.edit_save = edit_save = QLineEdit()
        edit_save.setPlaceholderText(generate_slug(2))
        edit_save.setDisabled(True)
        hbox_misc.addWidget(check_save)
        hbox_misc.addWidget(edit_save, 1)
        hbox_misc.addStretch(2)

        vbox.addWidget(group_misc)

        # Action bar
        action_bar = QWidget()
        hbox_action = QHBoxLayout(action_bar)
        hbox_action.setContentsMargins(0, 0, 0, 0)
        self.btn_back = btn_back = QPushButton('Back')
        self.btn_review = btn_review = QPushButton('Review')
        self.btn_run = btn_run = QPushButton('Run Routine')

        cool_font = QFont()
        cool_font.setWeight(QFont.DemiBold)
        # cool_font.setPixelSize(16)

        btn_back.setFixedSize(64, 64)
        btn_back.setFont(cool_font)
        btn_review.setFixedSize(64, 64)
        btn_review.setFont(cool_font)
        btn_run.setFixedSize(256, 64)
        btn_run.setFont(cool_font)
        hbox_action.addWidget(btn_back)
        hbox_action.addWidget(btn_review)
        hbox_action.addStretch(1)
        hbox_action.addWidget(btn_run)

        # vbox.addSpacing(16)
        vbox.addWidget(action_bar)

    def config_logic(self):
        self.btn_back.clicked.connect(self.go_home)
        self.cb_algo.currentIndexChanged.connect(self.select_algo)
        self.check_use_script.stateChanged.connect(self.toggle_use_script)
        self.btn_edit_script.clicked.connect(self.edit_script)
        self.cb_env.currentIndexChanged.connect(self.select_env)
        self.btn_env_play.clicked.connect(self.open_playground)
        self.btn_all_var.clicked.connect(self.check_all_var)
        self.btn_un_all_var.clicked.connect(self.uncheck_all_var)
        self.btn_add_var.clicked.connect(self.add_var)
        self.btn_all_obj.clicked.connect(self.check_all_obj)
        self.btn_un_all_obj.clicked.connect(self.uncheck_all_obj)
        self.check_only_var.stateChanged.connect(self.toggle_check_only_var)
        self.check_only_obj.stateChanged.connect(self.toggle_check_only_obj)
        self.btn_add_con.clicked.connect(self.add_constraint)
        self.check_save.stateChanged.connect(self.toggle_save)
        self.btn_review.clicked.connect(self.review)
        self.btn_run.clicked.connect(self.run)

    def refresh_ui(self, routine):
        self.routine = routine  # save routine for future reference

        self.algos = list_algo()
        self.envs = list_env()
        # Clean up the constraints list
        self.list_con.clear()

        if routine is None:
            # Reset the algo and env configs
            self.cb_algo.setCurrentIndex(-1)
            self.cb_env.setCurrentIndex(-1)

            # Reset the routine configs check box status
            self.check_only_var.setChecked(False)
            self.check_only_obj.setChecked(False)

            # Reset the save settings
            name = generate_slug(2)
            self.edit_save.setText('')
            self.edit_save.setPlaceholderText(name)
            self.check_save.setChecked(False)

            return

        # Fill in the algo and env configs
        name_algo = routine['algo']
        idx_algo = self.algos.index(name_algo)
        self.cb_algo.setCurrentIndex(idx_algo)
        self.edit_algo.setPlainText(ystring(routine['algo_params']))
        try:
            self.script = routine['config']['script']
        except KeyError:
            self.script = None

        name_env = routine['env']
        idx_env = self.envs.index(name_env)
        self.cb_env.setCurrentIndex(idx_env)
        self.edit_env.setPlainText(ystring(routine['env_params']))

        # Config the vocs panel
        self.check_only_var.setChecked(True)
        variables = [next(iter(v)) for v in routine['config']['variables']]
        for i in range(self.list_var.count()):
            item = self.list_var.item(i)
            item_widget = self.list_var.itemWidget(item)
            var_name = item_widget.check_name.text()
            try:
                idx = variables.index(var_name)
                vrange = routine['config']['variables'][idx][var_name]
                item_widget.sb_lower.sb.setValue(vrange[0])
                item_widget.sb_upper.sb.setValue(vrange[1])
                item_widget.check_name.setChecked(True)
            except ValueError:
                item_widget.check_name.setChecked(False)

        self.check_only_obj.setChecked(True)
        objectives = [next(iter(v)) for v in routine['config']['objectives']]
        for i in range(self.list_obj.count()):
            item = self.list_obj.item(i)
            item_widget = self.list_obj.itemWidget(item)
            obj_name = item_widget.check_name.text()
            try:
                idx = objectives.index(obj_name)
                rule = routine['config']['objectives'][idx][obj_name]
                idx_rule = 0 if rule == 'MINIMIZE' else 1
                item_widget.cb_rule.setCurrentIndex(idx_rule)
                item_widget.check_name.setChecked(True)
            except ValueError:
                item_widget.check_name.setChecked(False)

        if routine['config']['constraints'] is not None:
            for i in range(len(routine['config']['constraints'])):
                name = next(iter(routine['config']['constraints'][i]))
                relation, thres = routine['config']['constraints'][i][name][:2]
                try:
                    routine['config']['constraints'][i][name][2]
                    critical = True
                except:
                    critical = False
                relation = ['GREATER_THAN', 'LESS_THAN',
                            'EQUAL_TO'].index(relation)
                self.add_constraint(name, relation, thres, critical)

        # Config the save settings
        name = routine['name']
        self.edit_save.setPlaceholderText(generate_slug(2))
        self.edit_save.setText(name)
        self.check_save.setChecked(False)

        if self.script:
            self.check_use_script.setChecked(True)

    def select_algo(self, i):
        # Reset the script
        self.script = ''
        self.check_use_script.setChecked(False)

        if i == -1:
            self.edit_algo.setPlainText('')
            return

        name = self.algos[i]
        try:
            _, configs = get_algo(name)
            self.edit_algo.setPlainText(ystring(configs['params']))
        except Exception as e:
            self.cb_algo.setCurrentIndex(-1)
            return QMessageBox.critical(self, 'Error!', str(e))

    def toggle_use_script(self):
        if self.check_use_script.isChecked():
            self.btn_edit_script.show()
            self.edit_algo.setReadOnly(True)
            self.refresh_params_algo()
        else:
            self.btn_edit_script.hide()
            self.edit_algo.setReadOnly(False)

    def edit_script(self):
        algo = self.cb_algo.currentText()
        dlg = BadgerEditScriptDialog(self, algo, self.script, self.script_updated)
        dlg.exec()

    def script_updated(self, text):
        self.script = text
        self.refresh_params_algo()

    def refresh_params_algo(self):
        if not self.script:
            return

        try:
            tmp = {}
            exec(self.script, tmp)
            try:
                tmp['generate']  # test if generate function is defined
            except Exception as e:
                QMessageBox.warning(self, 'Please define a valid generate function!', str(e))
                return

            env_params = load_config(self.edit_env.toPlainText())
            try:
                intf_name = self.configs['interface'][0]
            except KeyError:
                intf_name = None
            configs = {
                'params': env_params,
                'interface': [intf_name]
            }
            env = instantiate_env(self.env, configs)
            # Function generate comes from the script
            params_algo = tmp['generate'](env, None)
            self.edit_algo.setPlainText(ystring(params_algo))
        except Exception as e:
            QMessageBox.warning(self, 'Invalid script!', str(e))

    def select_env(self, i):
        if i == -1:
            self.edit_env.setPlainText('')
            self.list_var.clear()
            self.dict_var.clear()
            self.list_obj.clear()
            self.dict_obj.clear()
            self.configs = None
            self.env = None
            self.btn_add_con.setDisabled(True)
            self.btn_add_var.setDisabled(True)
            self.routine = None
            return

        name = self.envs[i]
        try:
            env, configs = get_env(name)
            self.configs = configs
            self.env = env
            self.btn_add_con.setDisabled(False)
            self.btn_add_var.setDisabled(False)
            if self.check_use_script.isChecked():
                self.refresh_params_algo()
        except Exception as e:
            self.configs = None
            self.env = None
            self.cb_env.setCurrentIndex(-1)
            self.btn_add_con.setDisabled(True)
            self.btn_add_var.setDisabled(True)
            self.routine = None
            return QMessageBox.critical(self, 'Error!', str(e))

        self.edit_env.setPlainText(ystring(configs['params']))

        self.list_var.clear()
        vars_env = configs['variables']
        vars_combine = [*vars_env]
        if self.routine:  # check for the temp variables in vocs
            vars_vocs = self.routine['config']['variables']
            var_names_vocs = [next(iter(var)) for var in vars_vocs]
            var_names_env = [next(iter(var)) for var in vars_env]
            env_instance = instantiate_env(env, configs)
            for name in var_names_vocs:
                if name in var_names_env:
                    continue

                _var = {}
                _var[name] = env_instance._get_vrange(name)
                vars_combine.append(_var)
        for var in vars_combine:
            name = next(iter(var))
            vrange = var[name]
            item = QListWidgetItem(self.list_var)
            var_item = variable_item(name, vrange, self.toggle_var)
            item.setSizeHint(var_item.sizeHint())
            self.list_var.addItem(item)
            self.list_var.setItemWidget(item, var_item)
            self.dict_var[name] = item

        self.list_obj.clear()
        for obj in configs['observations']:
            item = QListWidgetItem(self.list_obj)
            obj_item = objective_item(obj, 0, self.toggle_obj)
            item.setSizeHint(obj_item.sizeHint())
            self.list_obj.addItem(item)
            self.list_obj.setItemWidget(item, obj_item)
            self.dict_obj[obj] = item

        self.list_con.clear()
        self.routine = None

    def open_playground(self):
        pass

    def check_all_var(self):
        for i in range(self.list_var.count()):
            item = self.list_var.item(i)
            item_widget = self.list_var.itemWidget(item)
            item_widget.check_name.setChecked(True)

    def uncheck_all_var(self):
        for i in range(self.list_var.count()):
            item = self.list_var.item(i)
            item_widget = self.list_var.itemWidget(item)
            item_widget.check_name.setChecked(False)

    def add_var(self):
        # TODO: Use a cached env
        env_params = load_config(self.edit_env.toPlainText())
        try:
            intf_name = self.configs['interface'][0]
        except KeyError:
            intf_name = None
        configs = {
            'params': env_params,
            'interface': [intf_name]
        }

        dlg = BadgerVariableDialog(self, self.env, configs, self.add_var_to_list)
        dlg.exec()

    def add_var_to_list(self, name, min, max):
        # Check if already in the list
        ok = False
        try:
            self.dict_var[name]
        except:
            ok = True

        if not ok:
            QMessageBox.warning(self, 'Variable already exists!', f'Variable {name} already exists!')
            return 1

        item = QListWidgetItem(self.list_var)
        var_item = variable_item(name, [min, max], self.toggle_var)
        var_item.check_name.setChecked(True)
        item.setSizeHint(var_item.sizeHint())
        self.list_var.addItem(item)
        self.list_var.setItemWidget(item, var_item)
        self.dict_var[name] = item

        return 0

    def check_all_obj(self):
        for i in range(self.list_obj.count()):
            item = self.list_obj.item(i)
            item_widget = self.list_obj.itemWidget(item)
            item_widget.check_name.setChecked(True)

    def uncheck_all_obj(self):
        for i in range(self.list_obj.count()):
            item = self.list_obj.item(i)
            item_widget = self.list_obj.itemWidget(item)
            item_widget.check_name.setChecked(False)

    def toggle_check_only_var(self):
        if self.check_only_var.isChecked():
            for i in range(self.list_var.count()):
                item = self.list_var.item(i)
                item_widget = self.list_var.itemWidget(item)
                if not item_widget.check_name.isChecked():
                    item_widget.hide()
                    item.setSizeHint(QSize(0, 0))
        else:
            for i in range(self.list_var.count()):
                item = self.list_var.item(i)
                item_widget = self.list_var.itemWidget(item)
                item_widget.show()
                item.setSizeHint(item_widget.sizeHint())

    def toggle_var(self, name):
        if self.check_only_var.isChecked():
            item = self.dict_var[name]
            item_widget = self.list_var.itemWidget(item)
            if item_widget.check_name.isChecked():
                item_widget.show()
                item.setSizeHint(item_widget.sizeHint())
            else:
                item_widget.hide()
                item.setSizeHint(QSize(0, 0))

    def toggle_check_only_obj(self):
        if self.check_only_obj.isChecked():
            for i in range(self.list_obj.count()):
                item = self.list_obj.item(i)
                item_widget = self.list_obj.itemWidget(item)
                if not item_widget.check_name.isChecked():
                    item_widget.hide()
                    item.setSizeHint(QSize(0, 0))
        else:
            for i in range(self.list_obj.count()):
                item = self.list_obj.item(i)
                item_widget = self.list_obj.itemWidget(item)
                item_widget.show()
                item.setSizeHint(item_widget.sizeHint())

    def toggle_obj(self, name):
        if self.check_only_obj.isChecked():
            item = self.dict_obj[name]
            item_widget = self.list_obj.itemWidget(item)
            if item_widget.check_name.isChecked():
                item_widget.show()
                item.setSizeHint(item_widget.sizeHint())
            else:
                item_widget.hide()
                item.setSizeHint(QSize(0, 0))

    def toggle_save(self):
        if self.check_save.isChecked():
            self.edit_save.setDisabled(False)
        else:
            self.edit_save.setDisabled(True)

    def add_constraint(self, name=None, relation=0, threshold=0, critical=False):
        if self.configs is None:
            return

        options = self.configs['observations']
        item = QListWidgetItem(self.list_con)
        con_item = constraint_item(options,
                                   lambda: self.list_con.takeItem(
                                       self.list_con.row(item)),
                                   name, relation, threshold, critical)
        item.setSizeHint(con_item.sizeHint())
        self.list_con.addItem(item)
        self.list_con.setItemWidget(item, con_item)
        # self.dict_con[''] = item

    def _compose_routine(self):
        # Compose the routine
        name = self.edit_save.text() or self.edit_save.placeholderText()
        algo = self.cb_algo.currentText()
        env = self.cb_env.currentText()
        algo_params = load_config(self.edit_algo.toPlainText())
        env_params = load_config(self.edit_env.toPlainText())

        variables = []
        for i in range(self.list_var.count()):
            item = self.list_var.item(i)
            item_widget = self.list_var.itemWidget(item)
            if item_widget.check_name.isChecked():
                var_name = item_widget.check_name.text()
                lb = item_widget.sb_lower.sb.value()
                ub = item_widget.sb_upper.sb.value()
                _dict = {}
                _dict[var_name] = [lb, ub]
                variables.append(_dict)

        objectives = []
        for i in range(self.list_obj.count()):
            item = self.list_obj.item(i)
            item_widget = self.list_obj.itemWidget(item)
            if item_widget.check_name.isChecked():
                obj_name = item_widget.check_name.text()
                rule = item_widget.cb_rule.currentText()
                _dict = {}
                _dict[obj_name] = rule
                objectives.append(_dict)

        constraints = []
        for i in range(self.list_con.count()):
            item = self.list_con.item(i)
            item_widget = self.list_con.itemWidget(item)
            critical = item_widget.check_crit.isChecked()
            con_name = item_widget.cb_obs.currentText()
            relation = CONS_RELATION_DICT[item_widget.cb_rel.currentText()]
            value = item_widget.sb.value()
            _dict = {}
            _dict[con_name] = [relation, value]
            if critical:
                _dict[con_name].append('CRITICAL')
            constraints.append(_dict)

        configs = {
            'variables': variables,
            'objectives': objectives,
            'constraints': constraints,
        }
        if self.check_use_script.isChecked():
            configs['script'] = self.script

        routine = {
            'name': name,
            'algo': algo,
            'env': env,
            'algo_params': algo_params,
            'env_params': env_params,
            'env_vranges': config_list_to_dict(variables),
            'config': configs,
        }

        # Sanity check and config normalization
        try:
            routine = normalize_routine(routine)
        except Exception as e:
            raise e

        return routine

    def review(self):
        try:
            routine = self._compose_routine()
        except Exception as e:
            return QMessageBox.critical(self, 'Error!', str(e))

        dlg = BadgerReviewDialog(self, routine)
        dlg.exec()

    def run_routine(self, routine, save):
        self.monitor = BadgerOptMonitor(self, routine, save)
        self.monitor.show()
        self.monitor.start()

    def run(self):
        try:
            routine = self._compose_routine()
        except Exception as e:
            return QMessageBox.critical(self, 'Error!', str(e))

        save = self.check_save.isChecked()

        try:
            self.run_routine(routine, save)
        except Exception as e:
            # raise e
            QMessageBox.critical(self, 'Error!', str(e))



import numpy as np
from PyQt5.QtWidgets import QVBoxLayout, QHBoxLayout, QWidget, QPushButton
from PyQt5.QtWidgets import QMessageBox, QComboBox, QLabel, QStyledItemDelegate
from PyQt5.QtCore import pyqtSignal, QThreadPool, Qt
from PyQt5.QtGui import QFont
import pyqtgraph as pg
from ..components.routine_runner import BadgerRoutineRunner
# from ..utils import AURORA_PALETTE, FROST_PALETTE
from ...utils import norm
from ...logbook import send_to_logbook, BADGER_LOGBOOK_ROOT
from ...archive import archive_run, BADGER_ARCHIVE_ROOT


class BadgerOptMonitor(QWidget):
    sig_pause = pyqtSignal(bool)  # True: pause, False: resume
    sig_stop = pyqtSignal()

    def __init__(self, parent, routine, save):
        super().__init__()
        self._parent = parent  # used for aligning with the parent window
        self.setAttribute(Qt.WA_DeleteOnClose, True)

        self.routine = routine
        self.obj_names = [next(iter(d))
                          for d in self.routine['config']['objectives']]
        self.var_names = [next(iter(d))
                          for d in self.routine['config']['variables']]
        # For plot type switching
        self.x_plot_type = 0  # 0: raw, 1: normalized
        self.vranges = np.array([d[next(iter(d))]
                                 for d in routine['config']['variables']])
        if self.routine['config']['constraints']:
            self.con_names = [next(iter(d))
                              for d in self.routine['config']['constraints']]
        else:
            self.con_names = []
        self.save = save

        self.init_ui()
        self.config_logic()

    def init_ui(self):
        self.setWindowTitle('Opt Monitor')
        self.resize(720, 960)
        self.center()

        vbox = QVBoxLayout(self)

        # Config bar
        config_bar = QWidget()
        hbox_config = QHBoxLayout(config_bar)
        hbox_config.setContentsMargins(0, 0, 0, 0)
        label = QLabel('Evaluation History (X) Plot Type')
        self.cb_plot = cb_plot = QComboBox()
        cb_plot.setItemDelegate(QStyledItemDelegate())
        cb_plot.addItems(['Raw', 'Normalized'])
        hbox_config.addWidget(label)
        hbox_config.addWidget(cb_plot, 1)

        # Set up the monitor
        # Don't set show=True or there will be a blank window flashing once
        self.monitor = monitor = pg.GraphicsLayoutWidget()
        # monitor.ci.setBorder((50, 50, 100))
        # monitor.resize(1000, 600)
        pg.setConfigOptions(antialias=True)

        self.label = label = pg.LabelItem(justify='right')
        label.setText(self._make_label())
        monitor.addItem(label)

        monitor.nextRow()

        self.plot_obj = plot_obj = monitor.addPlot(
            title='Evaluation History (Y)')
        plot_obj.setLabel('left', 'objectives')
        plot_obj.setLabel('bottom', 'iterations')
        plot_obj.showGrid(x=True, y=True)
        leg_obj = plot_obj.addLegend()
        leg_obj.setBrush((50, 50, 100, 200))

        monitor.nextRow()

        if self.con_names:
            self.plot_con = plot_con = monitor.addPlot(
                title='Evaluation History (C)')
            plot_con.setLabel('left', 'constraints')
            plot_con.setLabel('bottom', 'iterations')
            plot_con.showGrid(x=True, y=True)
            leg_con = plot_con.addLegend()
            leg_con.setBrush((50, 50, 100, 200))

            monitor.nextRow()

        self.plot_var = plot_var = monitor.addPlot(
            title='Evaluation History (X)')
        plot_var.setLabel('left', 'variables')
        plot_var.setLabel('bottom', 'iterations')
        plot_var.showGrid(x=True, y=True)
        leg_var = plot_var.addLegend()
        leg_var.setBrush((50, 50, 100, 200))

        if self.con_names:
            plot_con.setXLink(plot_obj)
        plot_var.setXLink(plot_obj)

        self.colors = ['c', 'g', 'm', 'y', 'b', 'r', 'w']
        self.symbols = ['o', 't', 't1', 's', 'p', 'h', 'd']
        self.curves_var = []
        self.curves_obj = []
        self.curves_con = []

        for i, obj_name in enumerate(self.obj_names):
            color = self.colors[i % len(self.colors)]
            symbol = self.symbols[i % len(self.colors)]
            _curve = self.plot_obj.plot(pen=pg.mkPen(color, width=3),
                                        # symbol=symbol,
                                        name=obj_name)
            self.curves_obj.append(_curve)

        for i, var_name in enumerate(self.var_names):
            color = self.colors[i % len(self.colors)]
            symbol = self.symbols[i % len(self.colors)]
            _curve = self.plot_var.plot(pen=pg.mkPen(color, width=3),
                                        # symbol=symbol,
                                        name=var_name)
            self.curves_var.append(_curve)

        if self.con_names:
            for i, con_name in enumerate(self.con_names):
                color = self.colors[i % len(self.colors)]
                symbol = self.symbols[i % len(self.colors)]
                _curve = self.plot_con.plot(pen=pg.mkPen(color, width=3),
                                            # symbol=symbol,
                                            name=con_name)
                self.curves_con.append(_curve)

        self.ins_obj = pg.InfiniteLine(movable=True, angle=90, label=None,
                                       labelOpts={
                                           'position': 0.1,
                                           'color': (200, 200, 100),
                                           'fill': (200, 200, 200, 50),
                                           'movable': True})
        if self.con_names:
            self.ins_con = pg.InfiniteLine(movable=True, angle=90, label=None,
                                           labelOpts={
                                               'position': 0.1,
                                               'color': (200, 200, 100),
                                               'fill': (200, 200, 200, 50),
                                               'movable': True})
        self.ins_var = pg.InfiniteLine(movable=True, angle=90, label=None,
                                       labelOpts={
                                           'position': 0.1,
                                           'color': (200, 200, 100),
                                           'fill': (200, 200, 200, 50),
                                           'movable': True})
        plot_obj.addItem(self.ins_obj)
        if self.con_names:
            plot_con.addItem(self.ins_con)
        plot_var.addItem(self.ins_var)

        # Action bar
        action_bar = QWidget()
        hbox_action = QHBoxLayout(action_bar)
        hbox_action.setContentsMargins(0, 0, 0, 0)

        cool_font = QFont()
        cool_font.setWeight(QFont.DemiBold)
        # cool_font.setPixelSize(16)

        self.btn_back = btn_back = QPushButton('Close')
        btn_back.setFixedSize(64, 64)
        btn_back.setFont(cool_font)
        self.btn_log = btn_log = QPushButton('Logbook')
        btn_log.setFixedSize(64, 64)
        btn_log.setFont(cool_font)
        self.btn_reset = btn_reset = QPushButton('Reset')
        btn_reset.setDisabled(True)
        btn_reset.setFixedSize(64, 64)
        btn_reset.setFont(cool_font)
        self.btn_opt = btn_opt = QPushButton('Optimal')
        btn_opt.setFixedSize(64, 64)
        btn_opt.setFont(cool_font)
        self.btn_set = btn_set = QPushButton('Set')
        btn_set.setDisabled(True)
        btn_set.setFixedSize(64, 64)
        btn_set.setFont(cool_font)
        self.btn_ctrl = btn_ctrl = QPushButton('Pause')
        btn_ctrl.setFixedSize(64, 64)
        btn_ctrl.setFont(cool_font)
        self.btn_stop = btn_stop = QPushButton('Stop')
        btn_stop.setFixedSize(128, 64)
        btn_stop.setFont(cool_font)
        hbox_action.addWidget(btn_back)
        hbox_action.addStretch(1)
        hbox_action.addWidget(btn_log)
        hbox_action.addWidget(btn_reset)
        hbox_action.addWidget(btn_opt)
        hbox_action.addWidget(btn_set)
        hbox_action.addWidget(btn_ctrl)
        hbox_action.addWidget(btn_stop)

        vbox.addWidget(config_bar)
        vbox.addWidget(monitor)
        vbox.addWidget(action_bar)

    def config_logic(self):
        self.vars = []
        self.objs = []
        self.cons = []

        self.running = False

        # Sync the inspector lines
        self.ins_obj.sigDragged.connect(self.ins_obj_dragged)
        self.ins_obj.sigPositionChangeFinished.connect(self.ins_drag_done)
        if self.con_names:
            self.ins_con.sigDragged.connect(self.ins_con_dragged)
            self.ins_con.sigPositionChangeFinished.connect(self.ins_drag_done)
        self.ins_var.sigDragged.connect(self.ins_var_dragged)
        self.ins_var.sigPositionChangeFinished.connect(self.ins_drag_done)
        self.plot_obj.scene().sigMouseClicked.connect(self.on_mouse_click)
        # sigMouseReleased.connect(self.on_mouse_click)

        # Thread runner
        self.thread_pool = QThreadPool(self)

        # Create the routine runner
        self.routine_runner = routine_runner = BadgerRoutineRunner(
            self.routine, self.save)
        routine_runner.signals.env_ready.connect(self.env_ready)
        routine_runner.signals.finished.connect(self.routine_finished)
        routine_runner.signals.progress.connect(self.update)
        routine_runner.signals.error.connect(self.on_error)
        routine_runner.signals.info.connect(self.on_info)

        self.sig_pause.connect(routine_runner.ctrl_routine)
        self.sig_stop.connect(routine_runner.stop_routine)

        self.btn_back.clicked.connect(self.close)
        self.btn_log.clicked.connect(self.logbook)
        self.btn_reset.clicked.connect(self.reset_env)
        self.btn_opt.clicked.connect(self.jump_to_optimal)
        self.btn_set.clicked.connect(self.set_vars)
        self.btn_ctrl.clicked.connect(self.ctrl_routine)
        self.btn_stop.clicked.connect(self.stop_routine)

        # Visualization
        self.cb_plot.currentIndexChanged.connect(self.select_x_plot_type)

    def center(self):
        pass

    def start(self):
        self.running = True  # if a routine runner is working
        self.thread_pool.start(self.routine_runner)

    def is_critical(self, cons):
        if not self.con_names:
            return False, None

        constraints = self.routine['config']['constraints']
        for i, con_dict in enumerate(constraints):
            name = self.con_names[i]
            if len(con_dict[name]) != 3:
                continue

            value = cons[i]
            relation, thres = con_dict[name][:2]
            if relation == 'GREATER_THAN':
                if value <= thres:
                    return True, f'{name} (current value: {value:.4f}) is less than {thres}!'
            elif relation == 'LESS_THAN':
                if value >= thres:
                    return True, f'{name} (current value: {value:.4f}) is greater than {thres}!'
            else:
                if value != thres:
                    return True, f'{name} (current value: {value:.4f}) is not equal to {thres}!'

        return False, None

    def update(self, vars, objs, cons):
        self.vars.append(vars)
        self.objs.append(objs)
        self.cons.append(cons)

        for i in range(len(self.obj_names)):
            self.curves_obj[i].setData(np.array(self.objs)[:, i])

        if self.con_names:
            for i in range(len(self.con_names)):
                self.curves_con[i].setData(np.array(self.cons)[:, i])

        if self.x_plot_type:
            _vars = norm(np.array(self.vars), self.vranges[:, 0], self.vranges[:, 1])
        else:
            _vars = np.array(self.vars)
        for i in range(len(self.var_names)):
            self.curves_var[i].setData(_vars[:, i])

        # Check critical condition
        critical, msg = self.is_critical(cons)
        if not critical:
            return

        self.sig_pause.emit(True)
        self.btn_ctrl.setText('Resume')

        reply = QMessageBox.warning(self,
                                    'Run Paused',
                                    f'Critical constraint was violated: {msg}\nTerminate the run?',
                                    QMessageBox.Yes | QMessageBox.No, QMessageBox.Yes)
        if reply == QMessageBox.Yes:
            self.sig_stop.emit()

    def env_ready(self, init_vars):
        self.env = self.routine_runner.env
        self.init_vars = init_vars

    def routine_finished(self):
        self.running = False
        self.btn_ctrl.setDisabled(True)
        self.btn_stop.setDisabled(True)
        self.btn_reset.setDisabled(False)
        self.btn_set.setDisabled(False)
        try:
            archive_run(self.routine_runner.routine, self.routine_runner.data)
        except Exception as e:
            QMessageBox.critical(self, 'Archive failed!', str(e))

        QMessageBox.information(
            self, 'Success!', f'Run data archived to {BADGER_ARCHIVE_ROOT}')

    def on_error(self, error):
        QMessageBox.critical(self, 'Error!', str(error))

    def on_info(self, msg):
        QMessageBox.information(self, 'Info', msg)

    def logbook(self):
        try:
            routine = self.routine_runner.routine
            data = self.routine_runner.data
            send_to_logbook(routine, data, self.monitor)
        except Exception as e:
            QMessageBox.critical(self, 'Log failed!', str(e))

        QMessageBox.information(
            self, 'Success!', f'Log saved to {BADGER_LOGBOOK_ROOT}')

    def ctrl_routine(self):
        if self.btn_ctrl.text() == 'Pause':
            self.sig_pause.emit(True)
            self.btn_ctrl.setText('Resume')
        else:
            self.sig_pause.emit(False)
            self.btn_ctrl.setText('Pause')

    def stop_routine(self):
        self.sig_stop.emit()

    def ins_obj_dragged(self, ins_obj):
        self.ins_var.setValue(ins_obj.value())
        if self.con_names:
            self.ins_con.setValue(ins_obj.value())

    def ins_con_dragged(self, ins_con):
        self.ins_var.setValue(ins_con.value())
        self.ins_obj.setValue(ins_con.value())

    def ins_var_dragged(self, ins_var):
        self.ins_obj.setValue(ins_var.value())
        if self.con_names:
            self.ins_con.setValue(ins_var.value())

    def ins_drag_done(self, ins):
        value = np.round(ins.value())
        self.ins_obj.setValue(value)
        if self.con_names:
            self.ins_con.setValue(value)
        self.ins_var.setValue(value)
        self.label.setText(self._make_label())

    def reset_env(self):
        reply = QMessageBox.question(self,
                                     'Reset Environment',
                                     f'Are you sure you want to reset the env vars back to {self.init_vars}?',
                                     QMessageBox.Yes | QMessageBox.No, QMessageBox.No)
        if reply != QMessageBox.Yes:
            return

        current_vars = self.env._get_vars(self.var_names)
        self.env._set_vars(self.var_names, self.init_vars)
        after_vars = self.env._get_vars(self.var_names)
        QMessageBox.information(
            self, 'Reset Environment', f'Env vars {current_vars} -> {after_vars}')

    def jump_to_optimal(self):
        pf = self.routine_runner.pf
        idx = pf.pareto_set[0][0]
        self.ins_obj.setValue(idx)
        if self.con_names:
            self.ins_con.setValue(idx)
        self.ins_var.setValue(idx)
        self.label.setText(self._make_label())

    def set_vars(self):
        df = self.routine_runner.data
        idx = int(self.ins_obj.value())
        solution = df.loc[idx, self.var_names].to_numpy()

        reply = QMessageBox.question(self,
                                     'Apply Solution',
                                     f'Are you sure you want to apply the current solution at {solution} to env?',
                                     QMessageBox.Yes | QMessageBox.No, QMessageBox.No)
        if reply != QMessageBox.Yes:
            return

        self.env._set_vars(self.var_names, solution)
        # center around the inspector
        self.plot_var.setXRange(idx - 3, idx + 3)
        # QMessageBox.information(
        #     self, 'Set Environment', f'Env vars have been set to {solution}')

    def select_x_plot_type(self, i):
        self.x_plot_type = i
        if i:
            _vars = norm(np.array(self.vars), self.vranges[:, 0], self.vranges[:, 1])
        else:
            _vars = np.array(self.vars)
        for i in range(len(self.var_names)):
            self.curves_var[i].setData(_vars[:, i])

    def _make_label(self):
        try:
            df = self.routine_runner.data
            idx = int(self.ins_obj.value())
            vars = df.loc[idx, self.var_names].to_numpy()
            objs = df.loc[idx, self.obj_names].to_numpy()
            cons = df.loc[idx, self.con_names].to_numpy()
            var_label = ', '.join(
                [f'{var_name}: {vars[i]:.4f}' for i, var_name in enumerate(self.var_names)])
            con_label = ', '.join(
                [f'{con_name}: {cons[i]:.4f}' for i, con_name in enumerate(self.con_names)])
            obj_label = ', '.join(
                [f'{obj_name}: {objs[i]:.4f}' for i, obj_name in enumerate(self.obj_names)])
        except:
            var_label = ', '.join(
                [f'{var_name}: {np.nan:.4f}' for i, var_name in enumerate(self.var_names)])
            con_label = ', '.join(
                [f'{con_name}: {np.nan:.4f}' for i, con_name in enumerate(self.con_names)])
            obj_label = ', '.join(
                [f'{obj_name}: {np.nan:.4f}' for i, obj_name in enumerate(self.obj_names)])

        return f"<span style='font-family: Courier; white-space: pre'>{var_label + ' | '}</span><span style='font-family: Courier; white-space: pre'>{obj_label + ' | '}</span><span style='font-family: Courier; white-space: pre'>{con_label}</span>"

    def on_mouse_click(self, event):
        # https://stackoverflow.com/a/64081483
        coor_obj = self.plot_obj.vb.mapSceneToView(event._scenePos)
        if self.con_names:
            coor_con = self.plot_con.vb.mapSceneToView(event._scenePos)
        coor_var = self.plot_var.vb.mapSceneToView(event._scenePos)

        flag = self.plot_obj.viewRect().contains(coor_obj) or \
            self.plot_var.viewRect().contains(coor_var)
        if self.con_names:
            flag = flag or self.plot_con.viewRect().contains(coor_con)

        if flag:
            idx = int(np.round(coor_obj.x()))

            self.ins_obj.setValue(idx)
            if self.con_names:
                self.ins_con.setValue(idx)
            self.ins_var.setValue(idx)

            self.label.setText(self._make_label())

    def closeEvent(self, event):
        if not self.running:
            return

        reply = QMessageBox.question(self,
                                     'Window Close',
                                     'Closing this window will terminate the run, and the run data would NOT be archived! Proceed?',
                                     QMessageBox.Yes | QMessageBox.No, QMessageBox.No)

        if reply == QMessageBox.Yes:
            self.sig_stop.emit()
            event.accept()
        else:
            event.ignore()



import os
import numpy as np
from PyQt5.QtWidgets import QHBoxLayout, QTabWidget, QVBoxLayout, QMessageBox, QTextEdit
from PyQt5.QtWidgets import QWidget, QSplitter, QPushButton, QTreeWidget, QTreeWidgetItem
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QFont
import pyqtgraph as pg
from ..windows.opt_monitor import BadgerOptMonitor
from ...archive import list_run, load_run, delete_run
from ...utils import ystring


class BadgerRunPage(QWidget):
    def __init__(self, go_routine=None, go_home=None):
        super().__init__()

        # go_xxx is a function that jumps to the corresponding page once called
        self.go_routine = go_routine
        self.go_home = go_home

        self.init_ui()
        self.config_logic()

    def init_ui(self):
        self.runs = runs = list_run()

        # Set up the layout
        vbox = QVBoxLayout(self)

        # History run list
        self.run_tree = run_tree = QTreeWidget()
        self.recent_item = None
        run_tree.setColumnCount(1)
        run_tree.setHeaderLabels(['History Run'])
        # run_list.setSpacing(1)
        items = []
        for year, dict_year in runs.items():
            item_year = QTreeWidgetItem([year])
            for month, dict_month in dict_year.items():
                item_month = QTreeWidgetItem([month])
                for day, list_day in dict_month.items():
                    item_day = QTreeWidgetItem([day])
                    for i, file in enumerate(list_day):
                        name = os.path.splitext(file)[0]
                        item_file = QTreeWidgetItem([name])
                        if not self.recent_item:
                            self.recent_item = item_file
                        item_day.addChild(item_file)
                    item_month.addChild(item_day)
                item_year.addChild(item_month)
            items.append(item_year)
        run_tree.insertTopLevelItems(0, items)

        splitter = QSplitter(Qt.Horizontal)

        self.run_tab = run_tab = QTabWidget()
        # Config the plot
        self.run_view = run_view = pg.GraphicsLayoutWidget()
        pg.setConfigOptions(antialias=True)
        self.plot_obj = plot_obj = run_view.addPlot(
            title='Evaluation History (Y)')
        plot_obj.setLabel('left', 'objectives')
        plot_obj.setLabel('bottom', 'iterations')
        plot_obj.showGrid(x=True, y=True)
        leg_obj = plot_obj.addLegend()
        leg_obj.setBrush((50, 50, 100, 200))

        run_view.nextRow()
        run_view.nextRow()

        self.plot_var = plot_var = run_view.addPlot(
            title='Evaluation History (X)')
        plot_var.setLabel('left', 'variables')
        plot_var.setLabel('bottom', 'iterations')
        plot_var.showGrid(x=True, y=True)
        leg_var = plot_var.addLegend()
        leg_var.setBrush((50, 50, 100, 200))

        plot_var.setXLink(plot_obj)

        # Config the raw data viewer
        self.run_edit = run_edit = QTextEdit()
        run_tab.addTab(run_view, 'Visualization')
        run_tab.addTab(run_edit, 'Raw Data')
        splitter.addWidget(run_tree)
        splitter.addWidget(run_tab)
        # splitter.setSizes([100, 200])
        splitter.setStretchFactor(0, 0)
        splitter.setStretchFactor(1, 1)

        vbox.addWidget(splitter, 1)

        # Action bar
        action_bar = QWidget()
        hbox_action = QHBoxLayout(action_bar)
        hbox_action.setContentsMargins(0, 0, 0, 0)

        cool_font = QFont()
        cool_font.setWeight(QFont.DemiBold)
        # cool_font.setPixelSize(16)

        self.btn_back = btn_back = QPushButton('Back')
        btn_back.setFixedSize(64, 64)
        btn_back.setFont(cool_font)
        self.btn_del = btn_del = QPushButton('Delete')
        btn_del.setFixedSize(64, 64)
        btn_del.setFont(cool_font)
        self.btn_run = btn_run = QPushButton('Rerun')
        btn_run.setFixedSize(64, 64)
        btn_run.setFont(cool_font)
        self.btn_load = btn_load = QPushButton('Load Routine')
        btn_load.setFixedSize(128, 64)
        btn_load.setFont(cool_font)
        hbox_action.addWidget(btn_back)
        hbox_action.addStretch(1)
        hbox_action.addWidget(btn_del)
        hbox_action.addWidget(btn_run)
        hbox_action.addWidget(btn_load)

        vbox.addWidget(action_bar)

    def refresh_ui(self):
        self.runs = runs = list_run()

        self.run_tree.clear()
        self.recent_item = None

        items = []
        for year, dict_year in runs.items():
            item_year = QTreeWidgetItem([year])
            for month, dict_month in dict_year.items():
                item_month = QTreeWidgetItem([month])
                for day, list_day in dict_month.items():
                    item_day = QTreeWidgetItem([day])
                    for file in list_day:
                        name = os.path.splitext(file)[0]
                        item_file = QTreeWidgetItem([name])
                        if not self.recent_item:
                            self.recent_item = item_file
                        item_day.addChild(item_file)
                    item_month.addChild(item_day)
                item_year.addChild(item_month)
            items.append(item_year)
        self.run_tree.insertTopLevelItems(0, items)

    def config_logic(self):
        self.colors = ['c', 'g', 'm', 'y', 'b', 'r', 'w']
        self.symbols = ['o', 't', 't1', 's', 'p', 'h', 'd']

        self.run_tree.currentItemChanged.connect(self.load_run)
        self.run_tree.setCurrentItem(self.recent_item)

        self.btn_back.clicked.connect(self.go_home)
        self.btn_del.clicked.connect(self.delete_run)
        self.btn_run.clicked.connect(self.rerun_routine)
        self.btn_load.clicked.connect(self.load_routine)

    def reconfig_logic(self):
        self.run_tree.setCurrentItem(self.recent_item)

    def load_run(self, current, previous):
        # self.current_item = current
        self.previous_item = previous

        try:
            self.run_name = run_name = current.text(0)
            self.run = run = load_run(run_name + '.yaml')
            self.run_edit.setText(ystring(run))
            self.plot_run()
        except:
            self.run_name = None
            self.run = None
            self.run_edit.setText('')
            self.plot_run()

    def plot_run(self):
        run = self.run
        self.plot_obj.clear()
        try:
            self.plot_con.clear()
        except:
            pass
        self.plot_var.clear()

        if not run:
            self.btn_del.setDisabled(True)
            self.btn_run.setDisabled(True)
            self.btn_load.setDisabled(True)

            try:
                self.run_view.removeItem(self.plot_con)
                del self.plot_con
            except:
                pass

            return

        self.btn_del.setDisabled(False)
        self.btn_run.setDisabled(False)
        self.btn_load.setDisabled(False)

        var_names = [next(iter(d))
                     for d in run['routine']['config']['variables']]
        obj_names = [next(iter(d))
                     for d in run['routine']['config']['objectives']]
        try:
            con_names = [next(iter(d))
                         for d in run['routine']['config']['constraints']]
        except:
            con_names = []
        data = run['data']

        for i, obj_name in enumerate(obj_names):
            color = self.colors[i % len(self.colors)]
            symbol = self.symbols[i % len(self.colors)]
            self.plot_obj.plot(np.array(data[obj_name]), pen=pg.mkPen(color, width=3),
                               # symbol=symbol,
                               name=obj_name)

        if con_names:
            try:
                self.plot_con
            except:
                self.plot_con = plot_con = self.run_view.addPlot(
                    row=1, col=0, title='Evaluation History (C)')
                plot_con.setLabel('left', 'constraints')
                plot_con.setLabel('bottom', 'iterations')
                plot_con.showGrid(x=True, y=True)
                leg_con = plot_con.addLegend()
                leg_con.setBrush((50, 50, 100, 200))

                plot_con.setXLink(self.plot_obj)

            for i, con_name in enumerate(con_names):
                color = self.colors[i % len(self.colors)]
                symbol = self.symbols[i % len(self.colors)]
                try:
                    self.plot_con.plot(np.array(data[con_name]), pen=pg.mkPen(color, width=3),
                                       # symbol=symbol,
                                       name=con_name)
                except:  # Mal-format data
                    pass
        else:
            try:
                self.run_view.removeItem(self.plot_con)
                del self.plot_con
            except:
                pass

        for i, var_name in enumerate(var_names):
            color = self.colors[i % len(self.colors)]
            symbol = self.symbols[i % len(self.colors)]
            self.plot_var.plot(np.array(data[var_name]), pen=pg.mkPen(color, width=3),
                               # symbol=symbol,
                               name=var_name)

    def delete_run(self):
        reply = QMessageBox.question(self,
                                     'Delete Run',
                                     f'Are you sure you want to delete this run data?',
                                     QMessageBox.Yes | QMessageBox.No, QMessageBox.No)
        if reply != QMessageBox.Yes:
            return

        if self.run_name is None:
            return

        delete_run(self.run_name + '.yaml')
        self.run_tree.setCurrentItem(self.previous_item)
        root = self.run_tree.invisibleRootItem()
        (self.previous_item.parent() or root).removeChild(self.previous_item)

    def load_routine(self):
        routine = self.run['routine']

        self.go_routine(routine)

    def rerun_routine(self):
        routine = self.run['routine']
        try:
            self.monitor = BadgerOptMonitor(self, routine, False)
            self.monitor.show()
            self.monitor.start()
        except Exception as e:
            # raise e
            QMessageBox.critical(self, 'Error!', str(e))



from pkg_resources import get_distribution
from PyQt5.QtWidgets import QMainWindow, QStackedWidget, QDesktopWidget
# from PyQt5.QtWidgets import QMenuBar, QMenu
from ..pages.home_page import BadgerHomePage


class BadgerMainWindow(QMainWindow):
    def __init__(self):
        super().__init__()

        self.init_ui()
        self.config_logic()

    def init_ui(self):
        version = get_distribution('badger-opt').version
        self.setWindowTitle(f'Badger v{version}')
        self.resize(960, 720)
        self.center()

        # Add menu bar
        # menu_bar = self.menuBar()
        # edit_menu = menu_bar.addMenu('Edit')
        # edit_menu.addAction('New')

        # Add pages
        self.home_page = BadgerHomePage()

        self.stacks = stacks = QStackedWidget()
        stacks.addWidget(self.home_page)

        stacks.setCurrentIndex(0)

        self.setCentralWidget(self.stacks)

    def center(self):
        qr = self.frameGeometry()
        cp = QDesktopWidget().availableGeometry().center()

        qr.moveCenter(cp)
        self.move(qr.topLeft())

    def config_logic(self):
        pass
